from urllib import parse

from .grpc_handler import GrpcHandler
from ..exceptions import MilvusException, ParamError
from .types import CompactionState, CompactionPlans, Replica, BulkInsertState, ResourceGroupInfo
from ..settings import DefaultConfig as config
from ..decorators import deprecated

from .check import is_legal_host, is_legal_port


class Milvus:
    @deprecated
    def __init__(self, host=None, port=config.GRPC_PORT, uri=config.GRPC_URI, channel=None, **kwargs):
        self.address = self.__get_address(host, port, uri)
        self._handler = GrpcHandler(address=self.address, channel=channel, **kwargs)

        if kwargs.get("pre_ping", False) is True:
            self._handler._wait_for_channel_ready()

    def __get_address(self, host=None, port=config.GRPC_PORT, uri=config.GRPC_URI):
        if host is None and uri is None:
            raise ParamError(message='Host and uri cannot both be None')

        if host is None:
            try:
                parsed_uri = parse.urlparse(uri, "tcp")
            except (Exception) as e:
                raise ParamError(message=f"Illegal uri [{uri}]: {e}") from e

            host, port = parsed_uri.hostname, parsed_uri.port

        host, port = str(host), str(port)
        if not (is_legal_host(host) and is_legal_port(port)):
            raise ParamError(message=f"Illegal host [{host}] or port [{port}]")

        return f"{host}:{port}"

    def _connection(self):
        return self.handler

    @property
    def name(self):
        return self._name

    @property
    def handler(self):
        return self._handler

    def reset_password(self, user, old_password, new_password):
        self._handler.reset_password(user, old_password, new_password)

    def close(self):
        if self._handler is None:
            raise MilvusException(message="Closing on closed handler")
        self.handler.close()
        self._handler = None

    def create_collection(self, collection_name, fields, shards_num=2, timeout=None, **kwargs):
        """ Creates a collection.

        :param collection_name: The name of the collection. A collection name can only include
        numbers, letters, and underscores, and must not begin with a number.
        :type  collection_name: str

        :param fields: Field parameters.
        :type  fields: dict

            ` {"fields": [
                    {"field": "A", "type": DataType.INT32}
                    {"field": "B", "type": DataType.INT64},
                    {"field": "C", "type": DataType.FLOAT},
                    {"field": "Vec", "type": DataType.FLOAT_VECTOR,
                     "params": {"dim": 128}}
                ],
            "auto_id": True}`

        :param shards_num: How wide to scale collection. Corresponds to how many active datanodes
                        can be used on insert.
        :type shards_num: int

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :param kwargs:
            * *consistency_level* (``str/int``) --
            Which consistency level to use when searching in the collection. For details, see
            https://github.com/milvus-io/milvus/blob/master/docs/developer_guides/how-guarantee-ts-works.md.
            Note: this parameter can be overwritten by the same parameter specified in search.
            * *properties* (``dict``) --


        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.create_collection(collection_name, fields, shards_num=shards_num, timeout=timeout, **kwargs)

    def drop_collection(self, collection_name, timeout=None):
        """
        Delete a specified collection.

        :param collection_name: The name of the collection to delete.
        :type  collection_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.drop_collection(collection_name, timeout=timeout)

    def has_collection(self, collection_name, timeout=None):
        """
        Checks whether a specified collection exists.

        :param collection_name: The name of the collection to check.
        :type  collection_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: If specified collection exists
        :rtype: bool

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.has_collection(collection_name, timeout=timeout)

    def describe_collection(self, collection_name, timeout=None):
        """
        Returns the schema of specified collection.
        Example: {'collection_name': 'create_collection_eXgbpOtn', 'auto_id': True, 'description': '',
                 'fields': [{'field_id': 100, 'name': 'INT32', 'description': '', 'type': 4, 'params': {},
                 {'field_id': 101, 'name': 'FLOAT_VECTOR', 'description': '', 'type': 101,
                 'params': {'dim': '128'}}]}

        :param collection_name: The name of the collection to describe.
        :type  collection_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: The schema of collection to describe.
        :rtype: dict

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.describe_collection(collection_name, timeout=timeout)

    def load_collection(self, collection_name, replica_number=1, timeout=None, **kwargs):
        """
        Loads a specified collection from disk to memory.

        :param collection_name: The name of the collection to load.
        :type  collection_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :param replica_number: Number of replication in memory to load
        :type replica_number: int

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.load_collection(collection_name, replica_number, timeout=timeout, **kwargs)

    def release_collection(self, collection_name, timeout=None):
        """
        Clear collection data from memory.

        :param collection_name: The name of collection to release.
        :type  collection_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.release_collection(collection_name=collection_name, timeout=timeout)

    def get_collection_stats(self, collection_name, timeout=None, **kwargs):
        """
        Returns collection statistics information.
        Example: {"row_count": 10}

        :param collection_name: The name of collection.
        :type  collection_name: str.

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: statistics information
        :rtype: dict

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            stats = handler.get_collection_stats(collection_name, timeout=timeout, **kwargs)
            result = {stat.key: stat.value for stat in stats}
            result["row_count"] = int(result["row_count"])
            return result

    def list_collections(self, timeout=None):
        """
        Returns a list of all collection names.

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: List of collection names, return when operation is successful
        :rtype: list[str]

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.list_collections(timeout=timeout)

    def create_partition(self, collection_name, partition_name, timeout=None):
        """
        Creates a partition in a specified collection. You only need to import the
        parameters of partition_name to create a partition. A collection cannot hold
        partitions of the same tag, whilst you can insert the same tag in different collections.

        :param collection_name: The name of the collection to create partitions in.
        :type  collection_name: str

        :param partition_name: The tag name of the partition to create.
        :type  partition_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.create_partition(collection_name, partition_name, timeout=timeout)

    def drop_partition(self, collection_name, partition_name, timeout=None):
        """
        Deletes the specified partition in a collection. Note that the default partition
        '_default' is not permitted to delete. When a partition deleted, all data stored in it
        will be deleted.

        :param collection_name: The name of the collection to delete partitions from.
        :type  collection_name: str

        :param partition_name: The tag name of the partition to delete.
        :type  partition_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.drop_partition(collection_name, partition_name, timeout=timeout)

    def has_partition(self, collection_name, partition_name, timeout=None):
        """
        Checks if a specified partition exists in a collection.

        :param collection_name: The name of the collection to find the partition in.
        :type  collection_name: str

        :param partition_name: The tag name of the partition to check
        :type  partition_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: Whether a specified partition exists in a collection.
        :rtype: bool

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.has_partition(collection_name, partition_name, timeout=timeout)

    def load_partitions(self, collection_name, partition_names, replica_number=1, timeout=None):
        """
        Load specified partitions from disk to memory.

        :param collection_name: The collection name which partitions belong to.
        :type  collection_name: str

        :param partition_names: The specified partitions to load.
        :type  partition_names: list[str]

        :param replica_number: The replication numbers to load.
        :type  replica_number: int

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.load_partitions(collection_name=collection_name,
                                           partition_names=partition_names,
                                           replica_number=replica_number, timeout=timeout)

    def release_partitions(self, collection_name, partition_names, timeout=None):
        """
        Clear partitions data from memory.

        :param collection_name: The collection name which partitions belong to.
        :type  collection_name: str

        :param partition_names: The specified partition to release.
        :type  partition_names: list[str]

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.release_partitions(collection_name=collection_name,
                                              partition_names=partition_names, timeout=timeout)

    def list_partitions(self, collection_name, timeout=None):
        """
        Returns a list of all partition tags in a specified collection.

        :param collection_name: The name of the collection to retrieve partition tags from.
        :type  collection_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: A list of all partition tags in specified collection.
        :rtype: list[str]

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.list_partitions(collection_name, timeout=timeout)

    def get_partition_stats(self, collection_name, partition_name, timeout=None, **kwargs):
        """
        Returns partition statistics information.
        Example: {"row_count": 10}

        :param collection_name: The name of collection.
        :type  collection_name: str.

        :param partition_name: The name of partition.
        :type  partition_name: str.

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: statistics information
        :rtype: dict

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            stats = handler.get_partition_stats(collection_name, partition_name, timeout=timeout, **kwargs)
            result = {stat.key: stat.value for stat in stats}
            result["row_count"] = int(result["row_count"])
            return result

    def create_alias(self, collection_name, alias, timeout=None, **kwargs):
        """
        Specify alias for a collection.
        Alias cannot be duplicated, you can't assign same alias to different collections.
        But you can specify multiple aliases for a collection, for example:
            before create_alias("collection_1", "bob"):
                collection_1's aliases = ["tom"]
            after create_alias("collection_1", "bob"):
                collection_1's aliases = ["tom", "bob"]

        :param collection_name: The name of collection.
        :type  collection_name: str.

        :param alias: The alias of the collection.
        :type  alias: str.

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.create_alias(collection_name, alias, timeout=timeout, **kwargs)

    def drop_alias(self, alias, timeout=None, **kwargs):
        """
        Delete an alias.
        This api no need to specify collection name because the milvus server knows which collection it belongs.
        For example:
            before drop_alias("bob"):
                collection_1's aliases = ["tom", "bob"]
            after drop_alias("bob"):
                collection_1's aliases = ["tom"]

        :param alias: The alias to be deleted.
        :type  alias: str.

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.drop_alias(alias, timeout=timeout, **kwargs)

    def alter_alias(self, collection_name, alias, timeout=None, **kwargs):
        """
        Change alias of a collection to another collection. If the alias doesn't exist, the api will return error.
        Alias cannot be duplicated, you can't assign same alias to different collections.
        This api can change alias owner collection, for example:
            before alter_alias("collection_2", "bob"):
                collection_1's aliases = ["bob"]
                collection_2's aliases = []
            after alter_alias("collection_2", "bob"):
                collection_1's aliases = []
                collection_2's aliases = ["bob"]

        :param collection_name: The name of collection.
        :type  collection_name: str.

        :param alias: The new alias of the collection.
        :type  alias: str.

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.alter_alias(collection_name, alias, timeout=timeout, **kwargs)

    def create_index(self, collection_name, field_name, params, timeout=None, **kwargs):
        """
        Creates an index for a field in a specified collection. Milvus does not support creating multiple
        indexes for a field. In a scenario where the field already has an index, if you create another one,
        the server will replace the existing index files with the new ones.

        Note that you need to call load_collection() or load_partitions() to make the new index take effect
        on searching tasks.

        :param collection_name: The name of the collection to create field indexes.
        :type  collection_name: str

        :param field_name: The name of the field to create an index for.
        :type  field_name: str

        :param params: Indexing parameters.
        :type  params: dict
            There are examples of supported indexes:

            IVF_FLAT:
                ` {
                    "metric_type":"L2",
                    "index_type": "IVF_FLAT",
                    "params":{"nlist": 1024}
                }`

            IVF_PQ:
                `{
                    "metric_type": "L2",
                    "index_type": "IVF_PQ",
                    "params": {"nlist": 1024, "m": 8, "nbits": 8}
                }`

            IVF_SQ8:
                `{
                    "metric_type": "L2",
                    "index_type": "IVF_SQ8",
                    "params": {"nlist": 1024}
                }`

            BIN_IVF_FLAT:
                `{
                    "metric_type": "JACCARD",
                    "index_type": "BIN_IVF_FLAT",
                    "params": {"nlist": 1024}
                }`

            HNSW:
                `{
                    "metric_type": "L2",
                    "index_type": "HNSW",
                    "params": {"M": 48, "efConstruction": 50}
                }`

            RHNSW_FLAT:
                `{
                    "metric_type": "L2",
                    "index_type": "RHNSW_FLAT",
                    "params": {"M": 48, "efConstruction": 50}
                }`

            RHNSW_PQ:
                `{
                    "metric_type": "L2",
                    "index_type": "RHNSW_PQ",
                    "params": {"M": 48, "efConstruction": 50, "PQM": 8}
                }`

            RHNSW_SQ:
                `{
                    "metric_type": "L2",
                    "index_type": "RHNSW_SQ",
                    "params": {"M": 48, "efConstruction": 50}
                }`

            ANNOY:
                `{
                    "metric_type": "L2",
                    "index_type": "ANNOY",
                    "params": {"n_trees": 8}
                }`

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :param kwargs:
            * *_async* (``bool``) --
              Indicate if invoke asynchronously. When value is true, method returns a IndexFuture object;
              otherwise, method returns results from server.
            * *_callback* (``function``) --
              The callback function which is invoked after server response successfully. It only take
              effect when _async is set to True.
            * *index_name* (``str``) --
              The name of index which will be created. Then you can use the index name to check the state of index.
              If no index name is specified, default index name is used.

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.create_index(collection_name, field_name, params, timeout=timeout, **kwargs)

    def drop_index(self, collection_name, field_name, timeout=None):
        """
        Removes the index of a field in a specified collection.

        :param collection_name: The name of the collection to remove the field index from.
        :type  collection_name: str

        :param field_name: The name of the field to remove the index of.
        :type  field_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.drop_index(collection_name=collection_name,
                                      field_name=field_name, index_name="", timeout=timeout)

    def describe_index(self, collection_name, index_name="", timeout=None):
        """
        Returns the schema of index built on specified field.
        Example: {'index_type': 'FLAT', 'metric_type': 'L2', 'params': {'nlist': 128}}

        :param collection_name: The name of the collection which field belong to.
        :type  collection_name: str

        :param field_name: The name of field to describe.
        :type  field_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: the schema of index built on specified field.
        :rtype: dict

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.describe_index(collection_name, index_name, timeout=timeout)

    def insert(self, collection_name, entities, partition_name=None, timeout=None, **kwargs):
        """
        Inserts entities in a specified collection.

        :param collection_name: The name of the collection to insert entities in.
        :type  collection_name: str.

        :param entities: The entities to insert.
        :type  entities: list

        :param partition_name: The name of the partition to insert entities in. The default value is
         None. The server stores entities in the “_default” partition by default.
        :type  partition_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :param kwargs:
            * *_async* (``bool``) --
              Indicate if invoke asynchronously. When value is true, method returns a MutationFuture object;
              otherwise, method returns results from server.
            * *_callback* (``function``) --
              The callback function which is invoked after server response successfully. It only take
              effect when _async is set to True.

        :return: list of ids of the inserted vectors.
        :rtype: list[int]

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.batch_insert(collection_name, entities, partition_name, timeout=timeout, **kwargs)

    def delete(self, collection_name, expr, partition_name=None, timeout=None, **kwargs):
        """
        Delete entities with an expression condition.
        And return results to show which primary key is deleted successfully

        :param collection_name: Name of the collection to delete entities from
        :type  collection_name: str

        :param expr: The expression to specify entities to be deleted
        :type  expr: str

        :param partition_name: Name of partitions that contain entities
        :type  partition_name: str

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur
        :type  timeout: float

        :return: list of ids of the deleted vectors.
        :rtype: list

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.delete(collection_name, expr, partition_name, timeout=timeout, **kwargs)

    def flush(self, collection_names=None, timeout=None, **kwargs):
        """
        Internally, Milvus organizes data into segments, and indexes are built in a per-segment manner.
        By default, a segment will be sealed if it grows large enough (according to segment size configuration).
        If any index is specified on certain field, the index-creating task will be triggered automatically
        when a segment is sealed.

        The flush() call will seal all the growing segments immediately of the given collection,
        and force trigger the index-creating tasks.

        :param collection_names: The name of collection to flush.
        :type  collection_names: list[str]

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :param kwargs:
            * *_async* (``bool``) --
              Indicate if invoke asynchronously. When value is true, method returns a FlushFuture object;
              otherwise, method returns results from server.
            * *_callback* (``function``) --
              The callback function which is invoked after server response successfully. It only take
              effect when _async is set to True.

        :return: None
        :rtype: NoneType

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.flush(collection_names, timeout=timeout, **kwargs)

    def search(self, collection_name, data, anns_field, param, limit, expression=None, partition_names=None,
               output_fields=None, timeout=None, round_decimal=-1, **kwargs):
        """
        Searches a collection based on the given expression and returns query results.

        :param collection_name: The name of the collection to search.
        :type  collection_name: str
        :param data: The vectors of search data, the length of data is number of query (nq), the dim of every vector in
                     data must be equal to vector field's of collection.
        :type  data: list[list[float]]
        :param anns_field: The vector field used to search of collection.
        :type  anns_field: str
        :param param: The parameters of search, such as nprobe, etc.
        :type  param: dict
        :param limit: The max number of returned record, we also called this parameter as topk.
        :type  limit: int
        :param expression: The boolean expression used to filter attribute.
        :type  expression: str
        :param partition_names: The names of partitions to search.
        :type  partition_names: list[str]
        :param output_fields: The fields to return in the search result, not supported now.
        :type  output_fields: list[str]
        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float
        :param round_decimal: The specified number of decimal places of returned distance
        :type  round_decimal: int
        :param kwargs:
            * *_async* (``bool``) --
              Indicate if invoke asynchronously. When value is true, method returns a SearchFuture object;
              otherwise, method returns results from server.
            * *_callback* (``function``) --
              The callback function which is invoked after server response successfully. It only take
              effect when _async is set to True.
            * *consistency_level* (``str/int``) --
              Which consistency level to use when searching in the collection. For details, see
              https://github.com/milvus-io/milvus/blob/master/docs/developer_guides/how-guarantee-ts-works.md.
              Note: this parameter will overwrite the same parameter user specified when creating the collection,
              if no consistency level was specified, search will use the collection consistency level.
            * *guarantee_timestamp* (``int``) --
              This function instructs Milvus to see all operations performed before a provided timestamp. If no
              such timestamp is provided, then Milvus will search all operations performed to date.
              Note: only used in Customized consistency level.
            * *graceful_time* (``int``) --
              Only used in bounded consistency level. If graceful_time is set, PyMilvus will use current timestamp minus
              the graceful_time as the `guarantee_timestamp`. This option is 5s by default if not set.
            * *travel_timestamp* (``int``) --
              Users can specify a timestamp in a search to get results based on a data view
                        at a specified point in time.

        :return: Query result. QueryResult is iterable and is a 2d-array-like class, the first dimension is
                 the number of vectors to query (nq), the second dimension is the number of limit(topk).
        :rtype: QueryResult

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.search(collection_name, data, anns_field, param, limit, expression, partition_names,
                                  output_fields, round_decimal=round_decimal, timeout=timeout, **kwargs)

    def get_query_segment_info(self, collection_name, timeout=None, **kwargs):
        """
        Notifies Proxy to return segments information from query nodes.

        :param collection_name: The name of the collection to get segments info.
        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur.
        :type  timeout: float

        :return: QuerySegmentInfo:
            QuerySegmentInfo is the growing segments's information in query cluster.
        :rtype: QuerySegmentInfo
        """
        with self._connection() as handler:
            return handler.get_query_segment_info(collection_name, timeout=timeout, **kwargs)

    def load_collection_progress(self, collection_name, timeout=None):
        """ {
            'loading_progress': '100%',
            'num_loaded_partitions': 3,
            'not_loaded_partitions': [],
        }
        """
        with self._connection() as handler:
            return handler.load_collection_progress(collection_name, timeout=timeout)

    def load_partitions_progress(self, collection_name, partition_names, timeout=None):
        """ {
            'loading_progress': '100%',
            'num_loaded_partitions': 3,
            'not_loaded_partitions': [],
        }
        """
        with self._connection() as handler:
            return handler.load_partitions_progress(collection_name, partition_names, timeout=timeout)

    def wait_for_loading_collection_complete(self, collection_name, timeout=None):
        with self._connection() as handler:
            return handler.wait_for_loading_collection(collection_name, timeout=timeout)

    def wait_for_loading_partitions_complete(self, collection_name, partition_names, timeout=None):
        with self._connection() as handler:
            return handler.wait_for_loading_partitions(collection_name, partition_names, timeout=timeout)

    def get_index_build_progress(self, collection_name, index_name, timeout=None):
        with self._connection() as handler:
            return handler.get_index_build_progress(collection_name, index_name, timeout=timeout)

    def wait_for_creating_index(self, collection_name, index_name, timeout=None):
        with self._connection() as handler:
            return handler.wait_for_creating_index(collection_name, index_name, timeout=timeout)

    def dummy(self, request_type, timeout=None):
        with self._connection() as handler:
            return handler.dummy(request_type, timeout=timeout)

    def query(self, collection_name, expr, output_fields=None, partition_names=None, timeout=None, **kwargs):
        """
        Query with a set of criteria, and results in a list of records that match the query exactly.

        :param collection_name: Name of the collection to retrieve entities from
        :type  collection_name: str

        :param expr: The query expression
        :type  expr: str

        :param output_fields: A list of fields to return
        :type  output_fields: list[str]

        :param partition_names: Name of partitions that contain entities
        :type  partition_names: list[str]

        :param timeout: An optional duration of time in seconds to allow for the RPC. When timeout
                        is set to None, client waits until server response or error occur
        :type  timeout: float

        :return: A list that contains all results
        :rtype: list

        :param kwargs:
            * *consistency_level* (``str/int``) --
              Which consistency level to use during a query on the collection. For details, see
              https://github.com/milvus-io/milvus/blob/master/docs/developer_guides/how-guarantee-ts-works.md.
              Note: this parameter will overwrite the same parameter user specified when creating the collection,
              if no consistency level was specified, query will use the collection consistency level.
            * *guarantee_timestamp* (``int``) --
              This function instructs Milvus to see all operations performed before a provided timestamp. If no
              such timestamp is specified, Milvus queries all operations performed to date.
              Note: only used in Customized consistency level.
            * *graceful_time* (``int``) --
              Only used in bounded consistency level. If graceful_time is set, PyMilvus will use current timestamp minus
              the graceful_time as the `guarantee_timestamp`. This option is 5s by default if not set.
            * *travel_timestamp* (``int``) --
              Users can specify a timestamp in a search to get results based on a data view
                        at a specified point in time.

        :raises RpcError: If gRPC encounter an error
        :raises ParamError: If parameters are invalid
        :raises MilvusException: If the return result from server is not ok
        """
        with self._connection() as handler:
            return handler.query(collection_name, expr, output_fields, partition_names, timeout=timeout, **kwargs)

    def load_balance(self, collection_name: str, src_node_id, dst_node_ids, sealed_segment_ids, timeout=None, **kwargs):
        """
        Do load balancing operation from source query node to destination query node.
        :param collection_name: The collection to balance.
        :type  collection_name: str

        :param src_node_id: The source query node id to balance.
        :type  src_node_id: int

        :param dst_node_ids: The destination query node ids to balance.
        :type  dst_node_ids: list[int]

        :param sealed_segment_ids: Sealed segment ids to balance.
        :type  sealed_segment_ids: list[int]

        :param timeout: The timeout for this method, unit: second
        :type  timeout: int

        :raises MilvusException: If query nodes not exist.
        :raises MilvusException: If sealed segments not exist.
        """
        with self._connection() as handler:
            return handler.load_balance(collection_name, src_node_id, dst_node_ids, sealed_segment_ids,
                                        timeout=timeout, **kwargs)

    def compact(self, collection_name, timeout=None, **kwargs) -> int:
        """
        Do compaction for the collection.

        :param collection_name: The collection name to compact
        :type  collection_name: str

        :param timeout: The timeout for this method, unit: second
        :type  timeout: int

        :return: the compaction ID
        :rtype: int

        :raises MilvusException: If collection name not exist.
        """
        with self._connection() as handler:
            return handler.compact(collection_name, timeout=timeout, **kwargs)

    def get_compaction_state(self, compaction_id: int, timeout=None, **kwargs) -> CompactionState:
        """
        Get compaction states of a targeted compaction id

        :param compaction_id: the id returned by compact
        :type  compaction_id: int

        :param timeout: The timeout for this method, unit: second
        :type  timeout: int

        :return: the state of the compaction
        :rtype: CompactionState

        :raises MilvusException: If compaction_id doesn't exist.
        """

        with self._connection() as handler:
            return handler.get_compaction_state(compaction_id, timeout=timeout, **kwargs)

    def wait_for_compaction_completed(self, compaction_id: int, timeout=None, **kwargs) -> CompactionState:
        with self._connection() as handler:
            return handler.wait_for_compaction_completed(compaction_id, timeout=timeout, **kwargs)

    def get_compaction_plans(self, compaction_id: int, timeout=None, **kwargs) -> CompactionPlans:
        """
        Get compaction states of a targeted compaction id

        :param compaction_id: the id returned by compact
        :type  compaction_id: int

        :param timeout: The timeout for this method, unit: second
        :type  timeout: int

        :return: the state of the compaction
        :rtype: CompactionState

        :raises MilvusException: If compaction_id doesn't exist.
        """
        with self._connection() as handler:
            return handler.get_compaction_plans(compaction_id, timeout=timeout, **kwargs)

    def get_replicas(self, collection_name: str, timeout=None, **kwargs) -> Replica:
        """ Get replica infos of a collection

        :param collection_name: the name of the collection
        :type  collection_name: str

        :param timeout: The timeout for this method, unit: second
        :type  timeout: int

        :return: the replica info
        :rtype: Replica

        :raises MilvusException: If collection_name doesn't exist.
        """
        with self._connection() as handler:
            return handler.get_replicas(collection_name, timeout=timeout, **kwargs)

    def do_bulk_insert(self, collection_name: str, partition_name: str, files: list, timeout=None, **kwargs) -> int:
        """ do_bulk_insert inserts entities through files, currently supports row-based json file.
        User need to create the json file with a specified json format which is described in the official user guide.
        Let's say a collection has two fields: "id" and "vec"(dimension=8), the row-based json format is:
          {"rows": [
              {"id": "0", "vec": [0.190, 0.046, 0.143, 0.972, 0.592, 0.238, 0.266, 0.995]},
              {"id": "1", "vec": [0.149, 0.586, 0.012, 0.673, 0.588, 0.917, 0.949, 0.944]},
              ......
            ]
          }
        The json file must be uploaded to root path of MinIO/S3 storage which is accessed by milvus server.
        For example:
            the milvus.yml specify the MinIO/S3 storage bucketName as "a-bucket", user can upload his json file
             to a-bucket/xxx.json, then call do_bulk_insert(files=["a-bucket/xxx.json"])

        :param collection_name: the name of the collection
        :type  collection_name: str

        :param partition_name: the name of the partition
        :type  partition_name: str

        :param files: related path of the file to be imported. for row-based json file, only allow
                      one file each invocation.
        :type  files: list[str]

        :param timeout: The timeout for this method, unit: second
        :type  timeout: int

        :param kwargs: other infos

        :return: id of the task
        :rtype:  int

        :raises BaseException: If collection_name doesn't exist.
        :raises BaseException: If the files input is illegal.
        """
        with self._connection() as handler:
            return handler.do_bulk_insert(collection_name, partition_name, files, timeout=timeout, **kwargs)

    def get_bulk_insert_state(self, task_id, timeout=None, **kwargs) -> BulkInsertState:
        """get_bulk_insert_state returns state of a certain task_id

        :param task_id: the task id returned by bulk_insert
        :type  task_id: int

        :return: BulkInsertState
        :rtype:  BulkInsertState
        """
        with self._connection() as handler:
            return handler.get_bulk_insert_state(task_id, timeout=timeout, **kwargs)

    def list_bulk_insert_tasks(self, timeout=None, **kwargs) -> list:
        """list_bulk_insert_tasks lists all bulk load tasks

        :param limit: maximum number of tasks returned, list all tasks if the value is 0, else return the latest tasks
        :type  limit: int

        :param collection_name: target collection name, list all tasks if the name is empty
        :type  collection_name: str

        :return: list[BulkInsertState]
        :rtype:  list[BulkInsertState]

        """
        with self._connection() as handler:
            return handler.list_bulk_insert_tasks(timeout=timeout, **kwargs)

    def create_user(self, user, password, timeout=None, **kwargs):
        """ Create a user using the given user and password.
        :param user: the user name.
        :type  user: str
        :param password: the password.
        :type  password: str
        :param timeout: The timeout for this method, unit: second
        :type  timeout: int
        """
        with self._connection() as handler:
            handler.create_user(user, password, timeout=timeout, **kwargs)

    def update_password(self, user, old_password, new_password, timeout=None, **kwargs):
        """
            Update the user password using the given user and password.
            You must provide the original password to check if the operation is valid.
            Note: after this operation, PyMilvus won't change the related header of this connection.
            So if you update credential for this connection, the connection may be invalid.

        :param user: the user name.
        :type  user: str
        :param old_password: the original password.
        :type  old_password: str
        :param new_password: the newly password of this user.
        :type  new_password: str
        """
        with self._connection() as handler:
            handler.update_password(user, old_password, new_password, timeout=timeout, **kwargs)

    def delete_user(self, user, timeout=None, **kwargs):
        """ Delete user corresponding to the username.
        :param user: the user name.
        :type  user: str
        :param timeout: The timeout for this method, unit: second
        :type  timeout: int
        """
        with self._connection() as handler:
            handler.delete_user(user, timeout=timeout, **kwargs)

    def list_usernames(self, timeout=None, **kwargs):
        """ List all usernames.
        :param timeout: The timeout for this method, unit: second
        :type  timeout: int
        :return list of str:
            The usernames in Milvus instances.
        """
        with self._connection() as handler:
            return handler.list_usernames(timeout=timeout, **kwargs)

    def create_role(self, role_name, timeout=None, **kwargs):
        """ Create Role
        :param role_name: the role name.
        :type  role_name: str
        """
        with self._connection() as handler:
            handler.create_role(role_name, timeout=timeout, **kwargs)

    def drop_role(self, role_name, timeout=None, **kwargs):
        """ Drop Role
        :param role_name: role name.
        :type  role_name: str
        """
        with self._connection() as handler:
            handler.drop_role(role_name, timeout=timeout, **kwargs)

    def add_user_to_role(self, username, role_name, timeout=None, **kwargs):
        """ Add User To Role
        :param username: user name.
        :type  username: str
        :param role_name: role name.
        :type  role_name: str
        """
        with self._connection() as handler:
            handler.add_user_to_role(username, role_name, timeout=timeout, **kwargs)

    def remove_user_from_role(self, username, role_name, timeout=None, **kwargs):
        """ Remove User From Role
        :param username: user name.
        :type  username: str
        :param role_name: role name.
        :type  role_name: str
        """
        with self._connection() as handler:
            handler.remove_user_from_role(username, role_name, timeout=timeout, **kwargs)

    def select_one_role(self, role_name, include_user_info, timeout=None, **kwargs):
        """ Select One Role Info
        :param role_name: role name.
        :type  role_name: str
        :param include_user_info: whether to obtain the user information associated with the role
        :type  include_user_info: bool
        """
        with self._connection() as handler:
            handler.select_one_role(role_name, include_user_info, timeout=timeout, **kwargs)

    def select_all_role(self, include_user_info, timeout=None, **kwargs):
        """ Select All Role Info
        :param include_user_info: whether to obtain the user information associated with roles
        :type  include_user_info: bool
        """
        with self._connection() as handler:
            handler.select_all_role(include_user_info, timeout=timeout, **kwargs)

    def select_one_user(self, username, include_role_info, timeout=None, **kwargs):
        """ Select One User Info
        :param username: user name.
        :type  username: str
        :param include_role_info: whether to obtain the role information associated with the user
        :type  include_role_info: bool
        """
        with self._connection() as handler:
            handler.select_one_user(username, include_role_info, timeout=timeout, **kwargs)

    def select_all_user(self, include_role_info, timeout=None, **kwargs):
        """ Select All User Info
        :param include_role_info: whether to obtain the role information associated with users
        :type  include_role_info: bool
        """
        with self._connection() as handler:
            handler.select_all_role(include_role_info, timeout=timeout, **kwargs)

    def grant_privilege(self, role_name, object, object_name, privilege,
                        timeout=None, **kwargs):
        """ Grant Privilege
        :param role_name: role name.
        :type  role_name: str
        :param object: object that will be granted the privilege.
        :type  object: str
        :param object_name: identifies a specific resource name.
        :type  object_name: str
        :param privilege: privilege name.
        :type  privilege: str
        """
        with self._connection() as handler:
            handler.grant_privilege(role_name, object, object_name, privilege,
                                    timeout=timeout, **kwargs)

    def revoke_privilege(self, role_name, object, object_name, privilege,
                         timeout=None, **kwargs):
        """ Revoke Privilege
        :param role_name: role name.
        :type  role_name: str
        :param object: object that will be granted the privilege.
        :type  object: str
        :param object_name: identifies a specific resource name.
        :type  object_name: str
        :param privilege: privilege name.
        :type  privilege: str
        """
        with self._connection() as handler:
            handler.revoke_privilege(role_name, object, object_name, privilege,
                                     timeout=timeout, **kwargs)

    def select_grant_for_one_role(self, role_name, timeout=None, **kwargs):
        """ Select the grant info about the role
        :param role_name: role name.
        :type  role_name: str
        """
        with self._connection() as handler:
            handler.select_grant_for_one_role(role_name, timeout=timeout, **kwargs)

    def select_grant_for_role_and_object(self, role_name, object, object_name,
                                         timeout=None, **kwargs):
        """ Select the grant info about the role and specific object
        :param role_name: role name.
        :type  role_name: str
        :param object: object that will be selected the privilege info.
        :type  object: str
        :param object_name: identifies a specific resource name.
        :type  object_name: str
        """
        with self._connection() as handler:
            handler.select_grant_for_role_and_object(role_name, object, object_name, timeout=timeout, **kwargs)

    def get_version(self, timeout=None, **kwargs):
        with self._connection() as handler:
            handler.get_version(timeout=timeout, **kwargs)

    def create_resource_group(self, name, timeout=None, **kwargs):
        """create resource group with specific name

        :param name: resource group name
        :type name: str
        """
        with self._connection() as handler:
            handler.create_resource_group(name, timeout=timeout, **kwargs)

    def drop_resource_group(self, name, timeout=None, **kwargs):
        """drop resource group with specific name

        :param name: resource group name
        :type name: str
        """
        with self._connection() as handler:
            handler.drop_resource_group(name, timeout=timeout, **kwargs)

    def list_resource_groups(self, timeout=None, **kwargs):
        """list all resource group names
        """
        with self._connection() as handler:
            handler.list_resource_groups(timeout=timeout, **kwargs)

    def describe_resource_group(self, name, timeout=None, **kwargs) -> ResourceGroupInfo:
        """describe resource group with specific name

        :param name: resource group info
        :type name: str
        :return: resource group info
        :rtype: ResourceGroupInfo
        """
        with self._connection() as handler:
            handler.describe_resource_group(name, timeout=timeout, **kwargs)

    def transfer_node(self, source, target, num_node, timeout=None, **kwargs):
        """transfer num_node from source resource group to target resource_group

        :param source: source resource group name
        :type source: str
        :param target: target resource group name
        :type target: str
        :param num_node: transfer node num
        :type num_node: int
        """
        with self._connection() as handler:
            handler.transfer_node(source, target, num_node, timeout=timeout, **kwargs)

    def transfer_replica(self, source, target, collection_name, num_replica, timeout=None, **kwargs):
        """transfer num_replica from source resource group to target resource group

        :param source: source resource group name
        :type source: str
        :param target: target resource group name
        :type target: str
        :param collection_name: collection name which replica belong to
        :type collection_name: str
        :param num_replica: transfer replica num
        :type num_replica: int
        """
        with self._connection() as handler:
            handler.transfer_replica(
                source, target, collection_name, num_replica, timeout=timeout, **kwargs)



import threading


class Singleton(type):
    _ins = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._ins:
            with cls._lock:
                if cls not in cls._ins:
                    ins = super().__call__(*args, **kwargs)
                    cls._ins[cls] = ins
        return cls._ins[cls]



import logging
import numpy
import pytest
from unittest import mock

from utils import *
from pymilvus import Collection, connections

LOGGER = logging.getLogger(__name__)


class TestCollections:

    #  @pytest.fixture(scope="function",)
    #  def collection(self):
    #      name = gen_collection_name()
    #      schema = gen_schema()
    #      yield Collection(name, schema=schema)
    #      if connections.get_connection().has_collection(name):
    #          connections.get_connection().drop_collection(name)

    def test_collection_by_DataFrame(self):
        from pymilvus import Collection
        from pymilvus import FieldSchema, CollectionSchema
        from pymilvus import DataType
        coll_name = gen_collection_name()
        fields = [
            FieldSchema("int64", DataType.INT64),
            FieldSchema("float", DataType.FLOAT),
            FieldSchema("float_vector", DataType.FLOAT_VECTOR, dim=128)
        ]

        prefix = "pymilvus.client.grpc_handler.GrpcHandler"

        collection_schema = CollectionSchema(fields, primary_field="int64")
        with mock.patch(f"{prefix}.__init__", return_value=None):
            with mock.patch(f"{prefix}._wait_for_channel_ready", return_value=None):
                connections.connect()

        with mock.patch(f"{prefix}.create_collection", return_value=None):
            with mock.patch(f"{prefix}.has_collection", return_value=False):
                collection = Collection(name=coll_name, schema=collection_schema)

        with mock.patch(f"{prefix}.create_collection", return_value=None):
            with mock.patch(f"{prefix}.has_collection", return_value=True):
                with mock.patch(f"{prefix}.describe_collection", return_value=collection_schema.to_dict()):
                    collection = Collection(name=coll_name)

        with mock.patch(f"{prefix}.drop_collection", return_value=None):
            with mock.patch(f"{prefix}.list_indexes", return_value=[]):
                with mock.patch(f"{prefix}.release_collection", return_value=None):
                    collection.drop()

        with mock.patch(f"{prefix}.close", return_value=None):
            connections.disconnect("default")

    @pytest.mark.xfail
    def test_constructor(self, collection):
        assert type(collection) is Collection

    @pytest.mark.xfail
    def test_construct_from_dataframe(self):
        assert type(Collection.construct_from_dataframe(gen_collection_name(), gen_pd_data(default_nb), primary_field="int64")[0]) is Collection

    @pytest.mark.xfail
    def test_schema(self, collection):
        schema = collection.schema
        description = "This is new description"
        with pytest.raises(AttributeError):
            schema.description = description
        with pytest.raises(AttributeError):
            collection.schema = schema

    @pytest.mark.xfail
    def test_description(self, collection):
        LOGGER.info(collection.description)
        description = "This is new description"
        with pytest.raises(AttributeError):
            collection.description = description

    @pytest.mark.xfail
    def test_name(self, collection):
        LOGGER.info(collection.name)
        with pytest.raises(AttributeError):
            collection.name = gen_collection_name()

    @pytest.mark.xfail
    def test_is_empty(self, collection):
        assert collection.is_empty is True

    @pytest.mark.xfail
    def test_num_entities(self, collection):
        assert collection.num_entities == 0

    @pytest.mark.xfail
    def test_drop(self, collection):
        collection.drop()

    @pytest.mark.xfail
    def test_load(self, collection):
        collection.load()

    @pytest.mark.xfail
    def test_release(self, collection):
        collection.release()

    @pytest.mark.xfail
    def test_insert(self, collection):
        data = gen_list_data(default_nb)
        collection.insert(data)

    @pytest.mark.xfail
    def test_insert_ret(self, collection):
        vectors = gen_vectors(1, default_dim, bool(0))
        data = [
            [1],
            [numpy.float32(1.0)],
            vectors
        ]
        result = collection.insert(data)
        print(result)
        assert "insert count" in str(result)
        assert "delete count" in str(result)
        assert "upsert count" in str(result)
        assert "timestamp" in str(result)

    @pytest.mark.xfail
    def test_search(self, collection):
        collection.search()

    @pytest.mark.xfail
    def test_get(self, collection):
        data = gen_list_data(default_nb)
        ids = collection.insert(data)
        assert len(ids) == default_nb
        res = collection.get(ids[0:10])

    @pytest.mark.xfail
    def test_query(self, collection):
        data = gen_list_data(default_nb)
        ids = collection.insert(data)
        assert len(ids) == default_nb
        ids_expr = ",".join(str(x) for x in ids)
        expr = "id in [ " + ids_expr + " ]"
        res = collection.query(expr)

    @pytest.mark.xfail
    def test_partitions(self, collection):
        assert len(collection.partitions) == 1

    @pytest.mark.xfail
    def test_partition(self, collection):
        collection.partition(gen_partition_name())

    @pytest.mark.xfail
    def test_has_partition(self, collection):
        assert collection.has_partition("_default") is True
        assert collection.has_partition(gen_partition_name()) is False

    @pytest.mark.xfail
    def test_drop_partition(self, collection):
        collection.drop_partition(gen_partition_name())

    @pytest.mark.xfail
    def test_indexes(self, collection):
        assert type(collection.indexes) is list
        assert len(collection.indexes) == 0

    @pytest.mark.xfail
    def test_index(self, collection):
        collection.index()

    @pytest.mark.xfail
    def test_create_index(self, collection, defa):
        collection.create_index(gen_field_name(), gen_index_name())

    @pytest.mark.xfail
    def test_has_index(self, collection):
        assert collection.has_index() is False

    @pytest.mark.xfail
    def test_drop_index(self, collection):
        collection.drop_index()

    def test_dummy(self):
        pass



from ..grpc_gen import common_pb2

ConsistencyLevel = common_pb2.ConsistencyLevel

LOGICAL_BITS = 18
LOGICAL_BITS_MASK = (1 << LOGICAL_BITS) - 1
EVENTUALLY_TS = 1
BOUNDED_TS = 2
DEFAULT_CONSISTENCY_LEVEL = ConsistencyLevel.Bounded



import logging
from mock_result import MockMutationResult


class MockMilvus:
    def __init__(self, host=None, port=None, handler="GRPC", pool="SingletonThread", **kwargs):
        self._collections = dict()
        self._collection_partitions = dict()
        self._collection_indexes = dict()

    def create_collection(self, collection_name, fields, timeout=None, **kwargs):
        if collection_name in self._collections:
            raise BaseException(1, f"Create collection failed: collection {collection_name} exist")
        self._collections[collection_name] = fields
        self._collection_partitions[collection_name] = {'_default'}
        self._collection_indexes[collection_name] = []
        logging.debug(f"create_collection: {collection_name}")

    def drop_collection(self, collection_name, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"describe collection failed: can't find collection: {collection_name}")
        self._collections.pop(collection_name)
        self._collection_partitions.pop(collection_name)
        logging.debug(f"drop_collection: {collection_name}")

    def has_collection(self, collection_name, timeout=None):
        logging.debug(f"has_collection: {collection_name}")
        return collection_name in self._collections

    def describe_collection(self, collection_name, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"describe collection failed: can't find collection: {collection_name}")
        logging.debug(f"describe_collection: {collection_name}")
        return self._collections[collection_name]

    def load_collection(self, collection_name, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"describe collection failed: can't find collection: {collection_name}")
        logging.debug(f"load_collection: {collection_name}")

    def release_collection(self, collection_name, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"describe collection failed: can't find collection: {collection_name}")
        logging.debug(f"release_collection: {collection_name}")

    def get_collection_stats(self, collection_name, timeout=None, **kwargs):
        if collection_name not in self._collections:
            raise BaseException(1, f"describe collection failed: can't find collection: {collection_name}")
        logging.debug(f"get_collection_stats: {collection_name}")
        return {'row_count': 0}

    def list_collections(self, timeout=None):
        logging.debug(f"list_collections")
        return list(self._collections.keys())

    def create_partition(self, collection_name, partition_tag, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"create partition failed: can't find collection: {collection_name}")
        if partition_tag in self._collection_partitions[collection_name]:
            raise BaseException(1, f"create partition failed: partition name = {partition_tag} already exists")
        logging.debug(f"create_partition: {collection_name}, {partition_tag}")
        self._collection_partitions[collection_name].add(partition_tag)

    def drop_partition(self, collection_name, partition_tag, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"DropPartition failed: can't find collection: {collection_name}")
        if partition_tag not in self._collection_partitions[collection_name]:
            raise BaseException(1, f"DropPartition failed: partition {partition_tag} does not exist")
        if partition_tag == "_default":
            raise BaseException(1, f"DropPartition failed: default partition cannot be deleted")
        logging.debug(f"drop_partition: {collection_name}, {partition_tag}")
        self._collection_partitions[collection_name].remove(partition_tag)

    def has_partition(self, collection_name, partition_tag, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"HasPartition failed: can't find collection: {collection_name}")
        logging.debug(f"has_partition: {collection_name}, {partition_tag}")
        return partition_tag in self._collection_partitions[collection_name]

    def load_partitions(self, collection_name, partition_names, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"describe collection failed: can't find collection: {collection_name}")
        for partition_name in partition_names:
            if partition_name not in self._collection_partitions[collection_name]:
                raise BaseException(1, f"partitionID of partitionName:{partition_name} can not be find")
        logging.debug(f"load_partition: {collection_name}, {partition_names}")

    def release_partitions(self, collection_name, partition_names, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"describe collection failed: can't find collection: {collection_name}")
        for partition_name in partition_names:
            if partition_name not in self._collection_partitions[collection_name]:
                raise BaseException(1, f"partitionID of partitionName:{partition_name} can not be find")
        logging.debug(f"release_partition: {collection_name}, {partition_names}")

    def get_partition_stats(self, collection_name, partition_name, timeout=None, **kwargs):
        if collection_name not in self._collections:
            raise BaseException(1, f"describe collection failed: can't find collection: {collection_name}")
        if partition_name not in self._collection_partitions[collection_name]:
            raise BaseException(1, f"GetPartitionStatistics failed: partition {partition_name} does not exist")
        logging.debug(f"get_partition_stats: {partition_name}")
        return {'row_count': 0}

    def list_partitions(self, collection_name, timeout=None):
        if collection_name not in self._collections:
            raise BaseException(1, f"can't find collection: {collection_name}")
        logging.debug(f"list_partitions: {collection_name}")
        return [e for e in self._collection_partitions[collection_name]]

    def create_index(self, collection_name, field_name, params, timeout=None, **kwargs):
        logging.debug(f"create_index: {collection_name}, {field_name}, {params}")
        index = {"field_name": field_name, "params": params}
        self._collection_indexes[collection_name].append(index)

    def drop_index(self, collection_name, field_name, timeout=None):
        logging.debug(f"drop_index: {collection_name}, {field_name}")
        self._collection_indexes[collection_name] = []

    def describe_index(self, collection_name, index_name="", timeout=None):
        logging.debug(f"describe_index: {collection_name}, {index_name}")
        if self._collection_indexes.get(collection_name) is None:
            return
        indexes = self._collection_indexes[collection_name].copy()
        if len(indexes) != 0:
            return indexes[0]

    def insert(self, collection_name, entities, ids=None, partition_tag=None, timeout=None, **kwargs):
        return MockMutationResult()

    def flush(self, collection_names=None, timeout=None, **kwargs):
        pass

    def search(self, collection_name, dsl, partition_tags=None, fields=None, timeout=None, **kwargs):
        pass

    def load_collection_progress(self, collection_name, timeout=None, **kwargs):
        return {'num_loaded_entities': 3000, 'num_total_entities': 5000}

    def load_partitions_progress(self, collection_name, partition_names, timeout=None, **kwargs):
        return {'num_loaded_entities': 3000, 'num_total_entities': 5000}

    def wait_for_loading_collection_complete(self, collection_name, timeout=None, **kwargs):
        pass

    def wait_for_loading_partitions_complete(self, collection_name, partition_names, timeout=None, **kwargs):
        pass

    def get_index_build_progress(self, collection_name, index_name, timeout=None, **kwargs):
        return {'total_rows': 5000, 'indexed_rows': 3000}

    def wait_for_creating_index(self, collection_name, index_name, timeout=None, **kwargs):
        return True, ""

    def close(self):
        pass


Milvus = MockMilvus



from pymilvus import utility, connections

_COLLECTION = "demo"

_HOST = '127.0.0.1'
_PORT = '19530'

_ROOT = "root"
_CONNECTION_NAME = "default"
_ANOTHER_CONNECTION_NAME = "another_conn"

_USER = "user"
_PASSWORD = "password"
_ANOTHER_USER = "another_user"
_ANOTHER_PASSWORD = "another_password"
_NEW_PASSWORD = "new_password"


def connect_without_auth(connection_name, host, port):
    print(f"connect to milvus without auth")
    print(f"connection: {connection_name}, host: {host}, port: {port}\n")
    connections.connect(alias=connection_name,
                        host=host,
                        port=port,
                        )


def connect_to_milvus(connection_name, host, port, user, password):
    print(f"connect to milvus user and password")
    print(f"connection: {connection_name}, host: {host}, port: {port}\n")
    connections.connect(alias=connection_name,
                        host=host,
                        port=port,
                        user=user,
                        password=password,
                        )


def create_user(connection_name, user, password):
    print(f"create user, connection: {connection_name}, user: {user}, password: {password}\n")
    utility.create_user(user, password, using=connection_name)


def update_password(connection_name, user, old_password, new_password):
    print(f"update password, connection: {connection_name}, user: {user}, old_password: {old_password}, new_password: {new_password}\n")
    utility.update_password(user, old_password, new_password, using=connection_name)


def delete_user(connection_name, user):
    print(f"delete user, connection: {connection_name}, user: {user}\n")
    utility.delete_user(user, using=connection_name)


def reset_password(connection_name, user, old_password, new_password):
    print(f"reset password, connection: {connection_name}, user: {user}, old_password: {old_password}, new_password: {new_password}\n")
    utility.reset_password(user, old_password, new_password, using=connection_name)


def test_connection(connection_name):
    print(f"test for {connection_name}")
    has = utility.has_collection(_COLLECTION, using=connection_name)
    print(f"has collection {_COLLECTION}: {has}")
    users = utility.list_usernames(using=connection_name)
    print(f"users in Milvus: {users}")
    print(f"test for {connection_name} done\n")


def run():
    connect_without_auth(_ROOT, _HOST, _PORT)
    test_connection(_ROOT)

    # after credential created, _ROOT was not able to call rpc of Milvus.
    # we should use new connection to communicate with Milvus.
    create_user(_ROOT, _USER, _PASSWORD)
    connect_to_milvus(_CONNECTION_NAME, _HOST, _PORT, _USER, _PASSWORD)
    test_connection(_CONNECTION_NAME)

    create_user(_CONNECTION_NAME, _ANOTHER_USER, _ANOTHER_PASSWORD)
    update_password(_CONNECTION_NAME, _ANOTHER_USER, _ANOTHER_PASSWORD, _NEW_PASSWORD)
    connect_to_milvus(_ANOTHER_CONNECTION_NAME, _HOST, _PORT, _ANOTHER_USER, _NEW_PASSWORD)
    test_connection(_ANOTHER_CONNECTION_NAME)

    reset_password(_CONNECTION_NAME, _USER, _PASSWORD, _NEW_PASSWORD)
    test_connection(_CONNECTION_NAME)

    delete_user(_CONNECTION_NAME, _ANOTHER_USER)
    delete_user(_CONNECTION_NAME, _USER)
    test_connection(_ROOT)


if __name__ == "__main__":
    run()




import datetime

from .types import DataType
from .constants import LOGICAL_BITS, LOGICAL_BITS_MASK
from ..exceptions import MilvusException

valid_index_types = [
    "FLAT",
    "IVF_FLAT",
    "IVF_SQ8",
    # "IVF_SQ8_HYBRID",
    "IVF_PQ",
    "HNSW",
    # "NSG",
    "ANNOY",
    "RHNSW_FLAT",
    "RHNSW_PQ",
    "RHNSW_SQ",
    "BIN_FLAT",
    "BIN_IVF_FLAT",
    "DISKANN",
    "AUTOINDEX"
]

valid_index_params_keys = [
    "nlist",
    "m",
    "nbits",
    "M",
    "efConstruction",
    "PQM",
    "n_trees"
]

valid_binary_index_types = [
    "BIN_FLAT",
    "BIN_IVF_FLAT"
]

valid_binary_metric_types = [
    "JACCARD",
    "HAMMING",
    "TANIMOTO",
    "SUBSTRUCTURE",
    "SUPERSTRUCTURE"
]


def hybridts_to_unixtime(ts):
    physical = ts >> LOGICAL_BITS
    return physical / 1000.0


def mkts_from_hybridts(hybridts, milliseconds=0., delta=None):
    if not isinstance(milliseconds, (int, float)):
        raise MilvusException(message="parameter milliseconds should be type of int or float")

    if isinstance(delta, datetime.timedelta):
        milliseconds += (delta.microseconds / 1000.0)
    elif delta is not None:
        raise MilvusException(message="parameter delta should be type of datetime.timedelta")

    if not isinstance(hybridts, int):
        raise MilvusException(message="parameter hybridts should be type of int")

    logical = hybridts & LOGICAL_BITS_MASK
    physical = hybridts >> LOGICAL_BITS

    new_ts = int((int((physical + milliseconds)) << LOGICAL_BITS) + logical)
    return new_ts


def mkts_from_unixtime(epoch, milliseconds=0., delta=None):
    if not isinstance(epoch, (int, float)):
        raise MilvusException(message="parameter epoch should be type of int or float")

    if not isinstance(milliseconds, (int, float)):
        raise MilvusException(message="parameter milliseconds should be type of int or float")

    if isinstance(delta, datetime.timedelta):
        milliseconds += (delta.microseconds / 1000.0)
    elif delta is not None:
        raise MilvusException(message="parameter delta should be type of datetime.timedelta")

    epoch += (milliseconds / 1000.0)
    int_msecs = int(epoch * 1000 // 1)
    return int(int_msecs << LOGICAL_BITS)


def mkts_from_datetime(d_time, milliseconds=0., delta=None):
    if not isinstance(d_time, datetime.datetime):
        raise MilvusException(message="parameter d_time should be type of datetime.datetime")

    return mkts_from_unixtime(d_time.timestamp(), milliseconds=milliseconds, delta=delta)


def check_invalid_binary_vector(entities) -> bool:
    for entity in entities:
        if entity['type'] == DataType.BINARY_VECTOR:
            if not isinstance(entity['values'], list) and len(entity['values']) == 0:
                return False

            dim = len(entity['values'][0]) * 8
            if dim == 0:
                return False

            for values in entity['values']:
                if len(values) * 8 != dim:
                    return False
                if not isinstance(values, bytes):
                    return False
    return True


def len_of(field_data) -> int:
    if field_data.HasField("scalars"):
        if field_data.scalars.HasField("bool_data"):
            return len(field_data.scalars.bool_data.data)

        if field_data.scalars.HasField("int_data"):
            return len(field_data.scalars.int_data.data)

        if field_data.scalars.HasField("long_data"):
            return len(field_data.scalars.long_data.data)

        if field_data.scalars.HasField("float_data"):
            return len(field_data.scalars.float_data.data)

        if field_data.scalars.HasField("double_data"):
            return len(field_data.scalars.double_data.data)

        if field_data.scalars.HasField("string_data"):
            return len(field_data.scalars.string_data.data)

        if field_data.scalars.HasField("bytes_data"):
            return len(field_data.scalars.bytes_data.data)

        raise MilvusException(message="Unsupported scalar type")

    if field_data.HasField("vectors"):
        dim = field_data.vectors.dim
        if field_data.vectors.HasField("float_vector"):
            total_len = len(field_data.vectors.float_vector.data)
            if total_len % dim != 0:
                raise MilvusException(message=f"Invalid vector length: total_len={total_len}, dim={dim}")
            return int(total_len / dim)

        total_len = len(field_data.vectors.binary_vector)
        return int(total_len / (dim / 8))

    raise MilvusException(message="Unknown data type")



import random

from pymilvus import Milvus, DataType

# This example shows how to:
#   1. connect to Milvus server
#   2. create a collection
#   3. insert entities
#   4. create index
#   5. search

_HOST = '127.0.0.1'
_PORT = '19530'

# Const names
_COLLECTION_NAME = 'demo'
_ID_FIELD_NAME = 'id_field'
_VECTOR_FIELD_NAME = 'float_vector_field'

# Vector parameters
_DIM = 128
_INDEX_FILE_SIZE = 32  # max file size of stored index

# Index parameters
_METRIC_TYPE = 'L2'
_INDEX_TYPE = 'IVF_FLAT'
_NLIST = 1024
_NPROBE = 16
_TOPK = 10

# Create milvus client instance
milvus = Milvus(_HOST, _PORT)


def create_collection(name):
    id_field = {
        "name": _ID_FIELD_NAME,
        "type": DataType.INT64,
        "auto_id": True,
        "is_primary": True,
    }
    vector_field = {
        "name": _VECTOR_FIELD_NAME,
        "type": DataType.FLOAT_VECTOR,
        "params": {"dim": _DIM},
    }
    fields = {"fields": [id_field, vector_field]}

    milvus.create_collection(collection_name=name, fields=fields)
    print("collection created:", name)


def drop_collection(name):
    if milvus.has_collection(name):
        milvus.drop_collection(name)
        print("collection dropped:", name)


def list_collections():
    collections = milvus.list_collections()
    print("list collection:")
    print(collections)


def get_collection_stats(name):
    stats = milvus.get_collection_stats(name)
    print("collection stats:")
    print(stats)


def insert(name, num, dim):
    vectors = [[random.random() for _ in range(dim)] for _ in range(num)]
    entities = [
        {"name": _VECTOR_FIELD_NAME, "type": DataType.FLOAT_VECTOR, "values": vectors},
    ]
    ids = milvus.insert(name, entities)
    return ids, vectors


def flush(name):
    milvus.flush([name])


def create_index(name, field_name):
    index_param = {
        "metric_type": _METRIC_TYPE,
        "index_type": _INDEX_TYPE,
        "params": {"nlist": _NLIST}
    }
    milvus.create_index(name, field_name, index_param)
    print("Create index: {}".format(index_param))


def drop_index(name, field_name):
    milvus.drop_index(name, field_name)
    print("Drop index:", field_name)


def load_collection(name):
    milvus.load_collection(name)


def release_collection(name):
    milvus.release_collection(name)


def search(name, vector_field, search_vectors, ids):
    nq = len(search_vectors)
    search_params = {"metric_type": _METRIC_TYPE, "params": {"nprobe": _NPROBE}}
    results = milvus.search(name, search_vectors, vector_field, param=search_params, limit=_TOPK)
    for i in range(nq):
        if results[i][0].distance == 0.0 or results[i][0].id == ids[0]:
            print("OK! search results: ", results[i][0].entity)
        else:
            print("FAIL! search results: ", results[i][0].entity)


def main():
    name = _COLLECTION_NAME
    vector_field = _VECTOR_FIELD_NAME

    drop_collection(name)
    create_collection(name)

    # show collections
    list_collections()

    # generate 10000 vectors with 128 dimension
    ids, vectors = insert(name, 10000, _DIM)

    # flush
    flush(name)

    # show row_count
    get_collection_stats(name)

    # create index
    create_index(name, vector_field)

    # load
    load_collection(name)

    # search
    search(name, vector_field, vectors[:10], ids)

    drop_index(name, vector_field)
    release_collection(name)
    drop_collection(name)


if __name__ == '__main__':
    main()



import threading
import datetime

from .singleton_utils import Singleton
from .utils import hybridts_to_unixtime
from .constants import EVENTUALLY_TS, BOUNDED_TS

from ..grpc_gen import common_pb2

ConsistencyLevel = common_pb2.ConsistencyLevel


class GTsDict(metaclass=Singleton):
    def __init__(self):
        # collection id -> last write ts
        self._last_write_ts_dict = {}
        self._last_write_ts_dict_lock = threading.Lock()

    def __repr__(self):
        return self._last_write_ts_dict.__repr__()

    def update(self, collection, ts):
        # use lru later if necessary
        with self._last_write_ts_dict_lock:
            if ts > self._last_write_ts_dict.get(collection, 0):
                self._last_write_ts_dict[collection] = ts

    def get(self, collection):
        return self._last_write_ts_dict.get(collection, 0)


# Return a GTsDict instance.
def _get_gts_dict():
    return GTsDict()


# Update the last write ts of collection.
def update_collection_ts(collection, ts):
    _get_gts_dict().update(collection, ts)


# Return a callback corresponding to the collection.
def update_ts_on_mutation(collection):
    def _update(mutation_result):
        update_collection_ts(collection, mutation_result.timestamp)

    return _update


# Get the last write ts of collection.
def get_collection_ts(collection):
    return _get_gts_dict().get(collection)


# Get the last write timestamp of collection.
def get_collection_timestamp(collection):
    ts = _get_gts_dict().get(collection)
    return hybridts_to_unixtime(ts)


# Get the last write datetime of collection.
def get_collection_datetime(collection, tz=None):
    timestamp = get_collection_timestamp(collection)
    return datetime.datetime.fromtimestamp(timestamp, tz=tz)


def get_eventually_ts():
    return EVENTUALLY_TS


def get_bounded_ts():
    return BOUNDED_TS


def construct_guarantee_ts(consistency_level, collection_name, kwargs):
    if consistency_level == ConsistencyLevel.Strong:
        # Milvus will assign a newest ts.
        kwargs["guarantee_timestamp"] = 0
    elif consistency_level == ConsistencyLevel.Session:
        # Using the last write ts of the collection.
        # TODO: get a timestamp from server?
        kwargs["guarantee_timestamp"] = get_collection_ts(collection_name) or get_eventually_ts()
    elif consistency_level == ConsistencyLevel.Bounded:
        # Milvus will assign ts according to the server timestamp and a configured time interval
        kwargs["guarantee_timestamp"] = get_bounded_ts()
    elif consistency_level == ConsistencyLevel.Eventually:
        # Using a very small timestamp.
        kwargs["guarantee_timestamp"] = get_eventually_ts()
    else:
        # Users customize the consistency level, no modification on `guarantee_timestamp`.
        pass



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('../..'))


# -- Project information -----------------------------------------------------

project = 'pymilvus'
copyright = '2019-2022, Milvus'
author = 'Milvus'

# The full version, including alpha/beta/rc tags
release = '2.1.0'
show_authors = True


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx_copybutton',
    'm2r',
    'sphinx.ext.autosummary',
    'sphinxcontrib.prettyspecialmethods',
    'sphinxcontrib.napoleon'
]


# Copy button configuration for python and shell
# regexp for:
# ">>> ", "... ", "$ ", "In [digit]", "(char) $ "
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: |\(.\) \$ "
copybutton_prompt_is_regexp = True

# Add any source suffix
source_suffix = ['.rst', '.md']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'python'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# napoleon styes
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_use_keyword = True
napoleon_custom_sections = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'
html_show_sphinx = False
copyright = '2019-2022 Zilliz. All rights reserved.'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for interceptors that operate on all RPC types."""

import collections

import grpc


class _GenericClientInterceptor(grpc.UnaryUnaryClientInterceptor,
                                grpc.UnaryStreamClientInterceptor,
                                grpc.StreamUnaryClientInterceptor,
                                grpc.StreamStreamClientInterceptor):

    def __init__(self, interceptor_function):
        super().__init__()
        self._fn = interceptor_function

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, False)
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response

    def intercept_unary_stream(self, continuation, client_call_details,
                               request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, True)
        response_it = continuation(new_details, next(new_request_iterator))
        return postprocess(response_it) if postprocess else response_it

    def intercept_stream_unary(self, continuation, client_call_details,
                               request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, False)
        response = continuation(new_details, new_request_iterator)
        return postprocess(response) if postprocess else response

    def intercept_stream_stream(self, continuation, client_call_details,
                                request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, True)
        response_it = continuation(new_details, new_request_iterator)
        return postprocess(response_it) if postprocess else response_it


def create(intercept_call):
    return _GenericClientInterceptor(intercept_call)


class _ClientCallDetails(
        collections.namedtuple(
            '_ClientCallDetails',
            ('method', 'timeout', 'metadata', 'credentials')),
        grpc.ClientCallDetails):
    pass


def header_adder_interceptor(header, value):

    def intercept_call(client_call_details, request_iterator, request_streaming,
                       response_streaming):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        metadata.append((
            header,
            value,
        ))
        client_call_details = _ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None

    return create(intercept_call)



# TODO(dragondriver): add more default configs to here
class DefaultConfigs:
    WaitTimeDurationWhenLoad = 0.5  # in seconds
    MaxVarCharLengthKey = "max_length"
    MaxVarCharLength = 65535
    EncodeProtocol = 'utf-8'
    IndexName = ""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x63ommon.proto\x12\x13milvus.proto.common\x1a google/protobuf/descriptor.proto\"L\n\x06Status\x12\x32\n\nerror_code\x18\x01 \x01(\x0e\x32\x1e.milvus.proto.common.ErrorCode\x12\x0e\n\x06reason\x18\x02 \x01(\t\"*\n\x0cKeyValuePair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"(\n\x0bKeyDataPair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"\x15\n\x04\x42lob\x12\r\n\x05value\x18\x01 \x01(\x0c\"c\n\x10PlaceholderValue\x12\x0b\n\x03tag\x18\x01 \x01(\t\x12\x32\n\x04type\x18\x02 \x01(\x0e\x32$.milvus.proto.common.PlaceholderType\x12\x0e\n\x06values\x18\x03 \x03(\x0c\"O\n\x10PlaceholderGroup\x12;\n\x0cplaceholders\x18\x01 \x03(\x0b\x32%.milvus.proto.common.PlaceholderValue\"#\n\x07\x41\x64\x64ress\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\"\x7f\n\x07MsgBase\x12.\n\x08msg_type\x18\x01 \x01(\x0e\x32\x1c.milvus.proto.common.MsgType\x12\r\n\x05msgID\x18\x02 \x01(\x03\x12\x11\n\ttimestamp\x18\x03 \x01(\x04\x12\x10\n\x08sourceID\x18\x04 \x01(\x03\x12\x10\n\x08targetID\x18\x05 \x01(\x03\"7\n\tMsgHeader\x12*\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x1c.milvus.proto.common.MsgBase\"M\n\x0c\x44MLMsgHeader\x12*\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x1c.milvus.proto.common.MsgBase\x12\x11\n\tshardName\x18\x02 \x01(\t\"\xbb\x01\n\x0cPrivilegeExt\x12\x34\n\x0bobject_type\x18\x01 \x01(\x0e\x32\x1f.milvus.proto.common.ObjectType\x12>\n\x10object_privilege\x18\x02 \x01(\x0e\x32$.milvus.proto.common.ObjectPrivilege\x12\x19\n\x11object_name_index\x18\x03 \x01(\x05\x12\x1a\n\x12object_name_indexs\x18\x04 \x01(\x05*\xd7\n\n\tErrorCode\x12\x0b\n\x07Success\x10\x00\x12\x13\n\x0fUnexpectedError\x10\x01\x12\x11\n\rConnectFailed\x10\x02\x12\x14\n\x10PermissionDenied\x10\x03\x12\x17\n\x13\x43ollectionNotExists\x10\x04\x12\x13\n\x0fIllegalArgument\x10\x05\x12\x14\n\x10IllegalDimension\x10\x07\x12\x14\n\x10IllegalIndexType\x10\x08\x12\x19\n\x15IllegalCollectionName\x10\t\x12\x0f\n\x0bIllegalTOPK\x10\n\x12\x14\n\x10IllegalRowRecord\x10\x0b\x12\x13\n\x0fIllegalVectorID\x10\x0c\x12\x17\n\x13IllegalSearchResult\x10\r\x12\x10\n\x0c\x46ileNotFound\x10\x0e\x12\x0e\n\nMetaFailed\x10\x0f\x12\x0f\n\x0b\x43\x61\x63heFailed\x10\x10\x12\x16\n\x12\x43\x61nnotCreateFolder\x10\x11\x12\x14\n\x10\x43\x61nnotCreateFile\x10\x12\x12\x16\n\x12\x43\x61nnotDeleteFolder\x10\x13\x12\x14\n\x10\x43\x61nnotDeleteFile\x10\x14\x12\x13\n\x0f\x42uildIndexError\x10\x15\x12\x10\n\x0cIllegalNLIST\x10\x16\x12\x15\n\x11IllegalMetricType\x10\x17\x12\x0f\n\x0bOutOfMemory\x10\x18\x12\x11\n\rIndexNotExist\x10\x19\x12\x13\n\x0f\x45mptyCollection\x10\x1a\x12\x1b\n\x17UpdateImportTaskFailure\x10\x1b\x12\x1a\n\x16\x43ollectionNameNotFound\x10\x1c\x12\x1b\n\x17\x43reateCredentialFailure\x10\x1d\x12\x1b\n\x17UpdateCredentialFailure\x10\x1e\x12\x1b\n\x17\x44\x65leteCredentialFailure\x10\x1f\x12\x18\n\x14GetCredentialFailure\x10 \x12\x18\n\x14ListCredUsersFailure\x10!\x12\x12\n\x0eGetUserFailure\x10\"\x12\x15\n\x11\x43reateRoleFailure\x10#\x12\x13\n\x0f\x44ropRoleFailure\x10$\x12\x1a\n\x16OperateUserRoleFailure\x10%\x12\x15\n\x11SelectRoleFailure\x10&\x12\x15\n\x11SelectUserFailure\x10\'\x12\x19\n\x15SelectResourceFailure\x10(\x12\x1b\n\x17OperatePrivilegeFailure\x10)\x12\x16\n\x12SelectGrantFailure\x10*\x12!\n\x1dRefreshPolicyInfoCacheFailure\x10+\x12\x15\n\x11ListPolicyFailure\x10,\x12\x12\n\x0eNotShardLeader\x10-\x12\x16\n\x12NoReplicaAvailable\x10.\x12\x13\n\x0fSegmentNotFound\x10/\x12\r\n\tForceDeny\x10\x30\x12\r\n\tRateLimit\x10\x31\x12\x12\n\x0eNodeIDNotMatch\x10\x32\x12\x14\n\x10UpsertAutoIDTrue\x10\x33\x12\x1c\n\x18InsufficientMemoryToLoad\x10\x34\x12\x18\n\x14MemoryQuotaExhausted\x10\x35\x12\x16\n\x12\x44iskQuotaExhausted\x10\x36\x12\x15\n\x11TimeTickLongDelay\x10\x37\x12\x11\n\rNotReadyServe\x10\x38\x12\x1b\n\x17NotReadyCoordActivating\x10\x39\x12\x12\n\x0eNotFoundTSafer\x10:\x12\x0f\n\x0b\x44\x61taCoordNA\x10\x64\x12\x12\n\rDDRequestRace\x10\xe8\x07*c\n\nIndexState\x12\x12\n\x0eIndexStateNone\x10\x00\x12\x0c\n\x08Unissued\x10\x01\x12\x0e\n\nInProgress\x10\x02\x12\x0c\n\x08\x46inished\x10\x03\x12\n\n\x06\x46\x61iled\x10\x04\x12\t\n\x05Retry\x10\x05*\x82\x01\n\x0cSegmentState\x12\x14\n\x10SegmentStateNone\x10\x00\x12\x0c\n\x08NotExist\x10\x01\x12\x0b\n\x07Growing\x10\x02\x12\n\n\x06Sealed\x10\x03\x12\x0b\n\x07\x46lushed\x10\x04\x12\x0c\n\x08\x46lushing\x10\x05\x12\x0b\n\x07\x44ropped\x10\x06\x12\r\n\tImporting\x10\x07*>\n\x0fPlaceholderType\x12\x08\n\x04None\x10\x00\x12\x10\n\x0c\x42inaryVector\x10\x64\x12\x0f\n\x0b\x46loatVector\x10\x65*\xb3\x0e\n\x07MsgType\x12\r\n\tUndefined\x10\x00\x12\x14\n\x10\x43reateCollection\x10\x64\x12\x12\n\x0e\x44ropCollection\x10\x65\x12\x11\n\rHasCollection\x10\x66\x12\x16\n\x12\x44\x65scribeCollection\x10g\x12\x13\n\x0fShowCollections\x10h\x12\x14\n\x10GetSystemConfigs\x10i\x12\x12\n\x0eLoadCollection\x10j\x12\x15\n\x11ReleaseCollection\x10k\x12\x0f\n\x0b\x43reateAlias\x10l\x12\r\n\tDropAlias\x10m\x12\x0e\n\nAlterAlias\x10n\x12\x13\n\x0f\x41lterCollection\x10o\x12\x14\n\x10RenameCollection\x10p\x12\x14\n\x0f\x43reatePartition\x10\xc8\x01\x12\x12\n\rDropPartition\x10\xc9\x01\x12\x11\n\x0cHasPartition\x10\xca\x01\x12\x16\n\x11\x44\x65scribePartition\x10\xcb\x01\x12\x13\n\x0eShowPartitions\x10\xcc\x01\x12\x13\n\x0eLoadPartitions\x10\xcd\x01\x12\x16\n\x11ReleasePartitions\x10\xce\x01\x12\x11\n\x0cShowSegments\x10\xfa\x01\x12\x14\n\x0f\x44\x65scribeSegment\x10\xfb\x01\x12\x11\n\x0cLoadSegments\x10\xfc\x01\x12\x14\n\x0fReleaseSegments\x10\xfd\x01\x12\x14\n\x0fHandoffSegments\x10\xfe\x01\x12\x18\n\x13LoadBalanceSegments\x10\xff\x01\x12\x15\n\x10\x44\x65scribeSegments\x10\x80\x02\x12\x10\n\x0b\x43reateIndex\x10\xac\x02\x12\x12\n\rDescribeIndex\x10\xad\x02\x12\x0e\n\tDropIndex\x10\xae\x02\x12\x0b\n\x06Insert\x10\x90\x03\x12\x0b\n\x06\x44\x65lete\x10\x91\x03\x12\n\n\x05\x46lush\x10\x92\x03\x12\x17\n\x12ResendSegmentStats\x10\x93\x03\x12\x0b\n\x06Search\x10\xf4\x03\x12\x11\n\x0cSearchResult\x10\xf5\x03\x12\x12\n\rGetIndexState\x10\xf6\x03\x12\x1a\n\x15GetIndexBuildProgress\x10\xf7\x03\x12\x1c\n\x17GetCollectionStatistics\x10\xf8\x03\x12\x1b\n\x16GetPartitionStatistics\x10\xf9\x03\x12\r\n\x08Retrieve\x10\xfa\x03\x12\x13\n\x0eRetrieveResult\x10\xfb\x03\x12\x14\n\x0fWatchDmChannels\x10\xfc\x03\x12\x15\n\x10RemoveDmChannels\x10\xfd\x03\x12\x17\n\x12WatchQueryChannels\x10\xfe\x03\x12\x18\n\x13RemoveQueryChannels\x10\xff\x03\x12\x1d\n\x18SealedSegmentsChangeInfo\x10\x80\x04\x12\x17\n\x12WatchDeltaChannels\x10\x81\x04\x12\x14\n\x0fGetShardLeaders\x10\x82\x04\x12\x10\n\x0bGetReplicas\x10\x83\x04\x12\x13\n\x0eUnsubDmChannel\x10\x84\x04\x12\x14\n\x0fGetDistribution\x10\x85\x04\x12\x15\n\x10SyncDistribution\x10\x86\x04\x12\x10\n\x0bSegmentInfo\x10\xd8\x04\x12\x0f\n\nSystemInfo\x10\xd9\x04\x12\x14\n\x0fGetRecoveryInfo\x10\xda\x04\x12\x14\n\x0fGetSegmentState\x10\xdb\x04\x12\r\n\x08TimeTick\x10\xb0\t\x12\x13\n\x0eQueryNodeStats\x10\xb1\t\x12\x0e\n\tLoadIndex\x10\xb2\t\x12\x0e\n\tRequestID\x10\xb3\t\x12\x0f\n\nRequestTSO\x10\xb4\t\x12\x14\n\x0f\x41llocateSegment\x10\xb5\t\x12\x16\n\x11SegmentStatistics\x10\xb6\t\x12\x15\n\x10SegmentFlushDone\x10\xb7\t\x12\x0f\n\nDataNodeTt\x10\xb8\t\x12\x15\n\x10\x43reateCredential\x10\xdc\x0b\x12\x12\n\rGetCredential\x10\xdd\x0b\x12\x15\n\x10\x44\x65leteCredential\x10\xde\x0b\x12\x15\n\x10UpdateCredential\x10\xdf\x0b\x12\x16\n\x11ListCredUsernames\x10\xe0\x0b\x12\x0f\n\nCreateRole\x10\xc0\x0c\x12\r\n\x08\x44ropRole\x10\xc1\x0c\x12\x14\n\x0fOperateUserRole\x10\xc2\x0c\x12\x0f\n\nSelectRole\x10\xc3\x0c\x12\x0f\n\nSelectUser\x10\xc4\x0c\x12\x13\n\x0eSelectResource\x10\xc5\x0c\x12\x15\n\x10OperatePrivilege\x10\xc6\x0c\x12\x10\n\x0bSelectGrant\x10\xc7\x0c\x12\x1b\n\x16RefreshPolicyInfoCache\x10\xc8\x0c\x12\x0f\n\nListPolicy\x10\xc9\x0c\x12\x18\n\x13\x43reateResourceGroup\x10\xa4\r\x12\x16\n\x11\x44ropResourceGroup\x10\xa5\r\x12\x17\n\x12ListResourceGroups\x10\xa6\r\x12\x1a\n\x15\x44\x65scribeResourceGroup\x10\xa7\r\x12\x11\n\x0cTransferNode\x10\xa8\r\x12\x14\n\x0fTransferReplica\x10\xa9\r*\"\n\x07\x44slType\x12\x07\n\x03\x44sl\x10\x00\x12\x0e\n\nBoolExprV1\x10\x01*B\n\x0f\x43ompactionState\x12\x11\n\rUndefiedState\x10\x00\x12\r\n\tExecuting\x10\x01\x12\r\n\tCompleted\x10\x02*X\n\x10\x43onsistencyLevel\x12\n\n\x06Strong\x10\x00\x12\x0b\n\x07Session\x10\x01\x12\x0b\n\x07\x42ounded\x10\x02\x12\x0e\n\nEventually\x10\x03\x12\x0e\n\nCustomized\x10\x04*\x8b\x01\n\x0bImportState\x12\x11\n\rImportPending\x10\x00\x12\x10\n\x0cImportFailed\x10\x01\x12\x11\n\rImportStarted\x10\x02\x12\x13\n\x0fImportPersisted\x10\x05\x12\x13\n\x0fImportCompleted\x10\x06\x12\x1a\n\x16ImportFailedAndCleaned\x10\x07*2\n\nObjectType\x12\x0e\n\nCollection\x10\x00\x12\n\n\x06Global\x10\x01\x12\x08\n\x04User\x10\x02*\x82\x07\n\x0fObjectPrivilege\x12\x10\n\x0cPrivilegeAll\x10\x00\x12\x1d\n\x19PrivilegeCreateCollection\x10\x01\x12\x1b\n\x17PrivilegeDropCollection\x10\x02\x12\x1f\n\x1bPrivilegeDescribeCollection\x10\x03\x12\x1c\n\x18PrivilegeShowCollections\x10\x04\x12\x11\n\rPrivilegeLoad\x10\x05\x12\x14\n\x10PrivilegeRelease\x10\x06\x12\x17\n\x13PrivilegeCompaction\x10\x07\x12\x13\n\x0fPrivilegeInsert\x10\x08\x12\x13\n\x0fPrivilegeDelete\x10\t\x12\x1a\n\x16PrivilegeGetStatistics\x10\n\x12\x18\n\x14PrivilegeCreateIndex\x10\x0b\x12\x18\n\x14PrivilegeIndexDetail\x10\x0c\x12\x16\n\x12PrivilegeDropIndex\x10\r\x12\x13\n\x0fPrivilegeSearch\x10\x0e\x12\x12\n\x0ePrivilegeFlush\x10\x0f\x12\x12\n\x0ePrivilegeQuery\x10\x10\x12\x18\n\x14PrivilegeLoadBalance\x10\x11\x12\x13\n\x0fPrivilegeImport\x10\x12\x12\x1c\n\x18PrivilegeCreateOwnership\x10\x13\x12\x17\n\x13PrivilegeUpdateUser\x10\x14\x12\x1a\n\x16PrivilegeDropOwnership\x10\x15\x12\x1c\n\x18PrivilegeSelectOwnership\x10\x16\x12\x1c\n\x18PrivilegeManageOwnership\x10\x17\x12\x17\n\x13PrivilegeSelectUser\x10\x18\x12 \n\x1cPrivilegeCreateResourceGroup\x10\x1a\x12\x1e\n\x1aPrivilegeDropResourceGroup\x10\x1b\x12\"\n\x1ePrivilegeDescribeResourceGroup\x10\x1c\x12\x1f\n\x1bPrivilegeListResourceGroups\x10\x1d\x12\x19\n\x15PrivilegeTransferNode\x10\x1e\x12\x1c\n\x18PrivilegeTransferReplica\x10\x1f\x12\x1f\n\x1bPrivilegeGetLoadingProgress\x10 \x12\x19\n\x15PrivilegeGetLoadState\x10!*S\n\tStateCode\x12\x10\n\x0cInitializing\x10\x00\x12\x0b\n\x07Healthy\x10\x01\x12\x0c\n\x08\x41\x62normal\x10\x02\x12\x0b\n\x07StandBy\x10\x03\x12\x0c\n\x08Stopping\x10\x04*c\n\tLoadState\x12\x15\n\x11LoadStateNotExist\x10\x00\x12\x14\n\x10LoadStateNotLoad\x10\x01\x12\x14\n\x10LoadStateLoading\x10\x02\x12\x13\n\x0fLoadStateLoaded\x10\x03:^\n\x11privilege_ext_obj\x12\x1f.google.protobuf.MessageOptions\x18\xe9\x07 \x01(\x0b\x32!.milvus.proto.common.PrivilegeExtBU\n\x0eio.milvus.grpcB\x0b\x43ommonProtoP\x01Z1github.com/milvus-io/milvus-proto/go-api/commonpb\xa0\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:
  google_dot_protobuf_dot_descriptor__pb2.MessageOptions.RegisterExtension(privilege_ext_obj)

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\016io.milvus.grpcB\013CommonProtoP\001Z1github.com/milvus-io/milvus-proto/go-api/commonpb\240\001\001'
  _ERRORCODE._serialized_start=933
  _ERRORCODE._serialized_end=2300
  _INDEXSTATE._serialized_start=2302
  _INDEXSTATE._serialized_end=2401
  _SEGMENTSTATE._serialized_start=2404
  _SEGMENTSTATE._serialized_end=2534
  _PLACEHOLDERTYPE._serialized_start=2536
  _PLACEHOLDERTYPE._serialized_end=2598
  _MSGTYPE._serialized_start=2601
  _MSGTYPE._serialized_end=4444
  _DSLTYPE._serialized_start=4446
  _DSLTYPE._serialized_end=4480
  _COMPACTIONSTATE._serialized_start=4482
  _COMPACTIONSTATE._serialized_end=4548
  _CONSISTENCYLEVEL._serialized_start=4550
  _CONSISTENCYLEVEL._serialized_end=4638
  _IMPORTSTATE._serialized_start=4641
  _IMPORTSTATE._serialized_end=4780
  _OBJECTTYPE._serialized_start=4782
  _OBJECTTYPE._serialized_end=4832
  _OBJECTPRIVILEGE._serialized_start=4835
  _OBJECTPRIVILEGE._serialized_end=5733
  _STATECODE._serialized_start=5735
  _STATECODE._serialized_end=5818
  _LOADSTATE._serialized_start=5820
  _LOADSTATE._serialized_end=5919
  _STATUS._serialized_start=71
  _STATUS._serialized_end=147
  _KEYVALUEPAIR._serialized_start=149
  _KEYVALUEPAIR._serialized_end=191
  _KEYDATAPAIR._serialized_start=193
  _KEYDATAPAIR._serialized_end=233
  _BLOB._serialized_start=235
  _BLOB._serialized_end=256
  _PLACEHOLDERVALUE._serialized_start=258
  _PLACEHOLDERVALUE._serialized_end=357
  _PLACEHOLDERGROUP._serialized_start=359
  _PLACEHOLDERGROUP._serialized_end=438
  _ADDRESS._serialized_start=440
  _ADDRESS._serialized_end=475
  _MSGBASE._serialized_start=477
  _MSGBASE._serialized_end=604
  _MSGHEADER._serialized_start=606
  _MSGHEADER._serialized_end=661
  _DMLMSGHEADER._serialized_start=663
  _DMLMSGHEADER._serialized_end=740
  _PRIVILEGEEXT._serialized_start=743
  _PRIVILEGEEXT._serialized_end=930
# @@protoc_insertion_point(module_scope)



import time
from enum import IntEnum
from ..grpc_gen import common_pb2
from ..exceptions import (
    AutoIDException,
    ExceptionsMessage,
    InvalidConsistencyLevel,
)
from ..grpc_gen import milvus_pb2 as milvus_types


ConsistencyLevel = common_pb2.ConsistencyLevel

class Status:
    """
    :attribute code: int (optional) default as ok

    :attribute message: str (optional) current status message
    """

    SUCCESS = common_pb2.Success
    UNEXPECTED_ERROR = common_pb2.UnexpectedError
    CONNECT_FAILED = 2
    PERMISSION_DENIED = 3
    COLLECTION_NOT_EXISTS = 4
    ILLEGAL_ARGUMENT = 5
    ILLEGAL_RANGE = 6
    ILLEGAL_DIMENSION = 7
    ILLEGAL_INDEX_TYPE = 8
    ILLEGAL_COLLECTION_NAME = 9
    ILLEGAL_TOPK = 10
    ILLEGAL_ROWRECORD = 11
    ILLEGAL_VECTOR_ID = 12
    ILLEGAL_SEARCH_RESULT = 13
    FILE_NOT_FOUND = 14
    META_FAILED = 15
    CACHE_FAILED = 16
    CANNOT_CREATE_FOLDER = 17
    CANNOT_CREATE_FILE = 18
    CANNOT_DELETE_FOLDER = 19
    CANNOT_DELETE_FILE = 20
    BUILD_INDEX_ERROR = 21
    ILLEGAL_NLIST = 22
    ILLEGAL_METRIC_TYPE = 23
    OUT_OF_MEMORY = 24
    INDEX_NOT_EXIST = 25
    EMPTY_COLLECTION = 26

    def __init__(self, code=SUCCESS, message="Success"):
        self.code = code
        self.message = message

    def __repr__(self):
        attr_list = [f'{key}={value}' for key, value in self.__dict__.items()]
        return f"{self.__class__.__name__}({', '.join(attr_list)})"

    def __eq__(self, other):
        """ Make Status comparable with self by code """
        if isinstance(other, int):
            return self.code == other

        return isinstance(other, self.__class__) and self.code == other.code

    def __ne__(self, other):
        return self != other

    def OK(self):
        return self.code == Status.SUCCESS


class DataType(IntEnum):
    NONE = 0
    BOOL = 1
    INT8 = 2
    INT16 = 3
    INT32 = 4
    INT64 = 5

    FLOAT = 10
    DOUBLE = 11

    STRING = 20
    VARCHAR = 21

    BINARY_VECTOR = 100
    FLOAT_VECTOR = 101

    UNKNOWN = 999


class RangeType(IntEnum):
    LT = 0  # less than
    LTE = 1  # less than or equal
    EQ = 2  # equal
    GT = 3  # greater than
    GTE = 4  # greater than or equal
    NE = 5  # not equal


class IndexType(IntEnum):
    INVALID = 0
    FLAT = 1
    IVFLAT = 2
    IVF_SQ8 = 3
    RNSG = 4
    IVF_SQ8H = 5
    IVF_PQ = 6
    HNSW = 11
    ANNOY = 12

    # alternative name
    IVF_FLAT = IVFLAT
    IVF_SQ8_H = IVF_SQ8H

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._name_}>"

    def __str__(self):
        return self._name_


class MetricType(IntEnum):
    INVALID = 0
    L2 = 1
    IP = 2
    # Only supported for byte vectors
    HAMMING = 3
    JACCARD = 4
    TANIMOTO = 5
    #
    SUBSTRUCTURE = 6
    SUPERSTRUCTURE = 7

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._name_}>"

    def __str__(self):
        return self._name_


class IndexState(IntEnum):
    IndexStateNone = 0
    Unissued = 1
    InProgress = 2
    Finished = 3
    Failed = 4
    Deleted = 5


class PlaceholderType(IntEnum):
    NoneType = 0
    BinaryVector = 100
    FloatVector = 101


class State(IntEnum):
    """
    UndefiedState:  Unknown
    Executing:      indicating this compaction has undone plans.
    Completed:      indicating all the plans of this compaction are done,
                    no matter successful or not.
    """

    UndefiedState = 0
    Executing = 1
    Completed = 2

    @staticmethod
    def new(s: int):
        if s == State.Executing:
            return State.Executing
        if s == State.Completed:
            return State.Completed
        return State.UndefiedState

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._name_}>"

    def __str__(self):
        return self._name_


class LoadState(IntEnum):
    """
    NotExist:     collection or partition isn't existed
    NotLoad:      collection or partition isn't loaded
    Loading:      collection or partition is loading
    Loaded:       collection or partition is loaded
    """

    NotExist = 0
    NotLoad = 1
    Loading = 2
    Loaded = 3

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._name_}>"

    def __str__(self):
        return self._name_

class CompactionState:
    """
    in_executing:   number of plans in executing
    in_timeout:     number of plans failed of timeout
    completed:      number of plans successfully completed
    """

    def __init__(self, compaction_id: int, state: State, in_executing: int, in_timeout: int, completed: int):
        self.compaction_id = compaction_id
        self.state = state
        self.in_executing = in_executing
        self.in_timeout = in_timeout
        self.completed = completed

    def __repr__(self):
        return f"""
CompactionState
 - compaction id: {self.compaction_id}
 - State: {self.state}
 - executing plan number: {self.in_executing}
 - timeout plan number: {self.in_timeout}
 - complete plan number: {self.completed}
"""


class Plan:
    def __init__(self, sources: list, target: int):
        self.sources = sources
        self.target = target

    def __repr__(self):
        return f"""
Plan:
 - sources: {self.sources}
 - target: {self.target}
"""


class CompactionPlans:
    def __init__(self, compaction_id: int, state: int):
        self.compaction_id = compaction_id
        self.state = State.new(state)
        self.plans = []

    def __repr__(self):
        return f"""
Compaction Plans:
 - compaction id: {self.compaction_id}
 - state: {self.state}
 - plans: {self.plans}
 """


def cmp_consistency_level(l1, l2):
    if isinstance(l1, str):
        try:
            l1 = ConsistencyLevel.Value(l1)
        except ValueError:
            return False

    if isinstance(l2, str):
        try:
            l2 = ConsistencyLevel.Value(l2)
        except ValueError:
            return False

    if isinstance(l1, int):
        if l1 not in ConsistencyLevel.values():
            return False

    if isinstance(l2, int):
        if l2 not in ConsistencyLevel.values():
            return False

    return l1 == l2


def get_consistency_level(consistency_level):
    if isinstance(consistency_level, int):
        if consistency_level in ConsistencyLevel.values():
            return consistency_level
        raise InvalidConsistencyLevel(message=f"invalid consistency level: {consistency_level}")
    if isinstance(consistency_level, str):
        try:
            return ConsistencyLevel.Value(consistency_level)
        except ValueError as e:
            raise InvalidConsistencyLevel(message=f"invalid consistency level: {consistency_level}") from e
    raise InvalidConsistencyLevel(message="invalid consistency level")


class Shard:
    def __init__(self, channel_name: str, shard_nodes: list, shard_leader: int):
        self._channel_name = channel_name
        self._shard_nodes = set(shard_nodes)
        self._shard_leader = shard_leader

    def __repr__(self):
        return f"""Shard: <channel_name:{self.channel_name}>, <shard_leader:{self.shard_leader}>, <shard_nodes:{self.shard_nodes}>"""

    @property
    def channel_name(self) -> str:
        return self._channel_name

    @property
    def shard_nodes(self):
        return self._shard_nodes

    @property
    def shard_leader(self) -> int:
        return self._shard_leader


class Group:
    def __init__(self, group_id: int, shards: list, group_nodes: list, resource_group: str, num_outbound_node: dict):
        self._id = group_id
        self._shards = shards
        self._group_nodes = tuple(group_nodes)
        self._resource_group = resource_group
        self._num_outbound_node = num_outbound_node

    def __repr__(self) -> str:
        s = f"Group: <group_id:{self.id}>, <group_nodes:{self.group_nodes}>, <shards:{self.shards}>, <resource_group: {self.resource_group}>, <num_outbound_node: {self.num_outbound_node}>"
        return s

    @property
    def id(self):
        return self._id

    @property
    def group_nodes(self):
        return self._group_nodes

    @property
    def shards(self):
        return self._shards

    @property
    def resource_group(self):
        return self._resource_group

    @property
    def num_outbound_node(self):
        return self._num_outbound_node

class Replica:
    """
    Replica groups:
        - Group: <group_id:2>, <group_nodes:(1, 2, 3)>,
            <shards:[Shard: <shard_id:10>, <channel_name:channel-1>, <shard_leader:1>, <shard_nodes:(1, 2, 3)>]>
        - Group: <group_id:2>, <group_nodes:(1, 2, 3)>,
            <shards:[Shard: <shard_id:10>, <channel_name:channel-1>, <shard_leader:1>, <shard_nodes:(1, 2, 3)>]>
    """

    def __init__(self, groups: list):
        self._groups = groups

    def __repr__(self) -> str:
        s = "Replica groups:"
        for g in self.groups:
            s += f"\n- {g}"
        return s

    @property
    def groups(self):
        return self._groups


class BulkInsertState:
    """enum states of bulk insert task"""
    ImportPending = 0
    ImportFailed = 1
    ImportStarted = 2
    ImportPersisted = 5
    ImportCompleted = 6
    ImportFailedAndCleaned = 7
    ImportUnknownState = 100

    """pre-defined keys of bulk insert task info"""
    FAILED_REASON = "failed_reason"
    IMPORT_FILES = "files"
    IMPORT_COLLECTION = "collection"
    IMPORT_PARTITION = "partition"
    IMPORT_PROGRESS = "progress_percent"

    """
    Bulk insert state example:
        - taskID            : 44353845454358,
        - state             : "BulkLoadPersisted",
        - row_count         : 1000,
        - infos             : {"files": "rows.json", "collection": "c1", "partition": "", "failed_reason": ""},
        - id_list           : [44353845455401, 44353845456401]
        - create_ts         : 1661398759,
    """

    state_2_state = {
        common_pb2.ImportPending: ImportPending,
        common_pb2.ImportFailed: ImportFailed,
        common_pb2.ImportStarted: ImportStarted,
        common_pb2.ImportPersisted: ImportPersisted,
        common_pb2.ImportCompleted: ImportCompleted,
        common_pb2.ImportFailedAndCleaned: ImportFailedAndCleaned,
    }

    state_2_name = {
        ImportPending: "Pending",
        ImportFailed: "Failed",
        ImportStarted: "Started",
        ImportPersisted: "Persisted",
        ImportCompleted: "Completed",
        ImportFailedAndCleaned: "Failed and cleaned",
        ImportUnknownState: "Unknown",
    }

    def __init__(self, task_id, state, row_count: int, id_ranges: list, infos, create_ts: int):
        self._task_id = task_id
        self._state = state
        self._row_count = row_count
        self._id_ranges = id_ranges
        self._create_ts = create_ts

        self._infos = {kv.key: kv.value for kv in infos}

    def __repr__(self) -> str:
        fmt = """<Bulk insert state:
    - taskID          : {},
    - state           : {},
    - row_count       : {},
    - infos           : {},
    - id_ranges       : {},
    - create_ts       : {}
>"""
        return fmt.format(self._task_id, self.state_name, self.row_count, self.infos,
                          self.id_ranges, self.create_time_str)

    @property
    def task_id(self):
        """
        Return unique id of this task.
        """
        return self._task_id

    @property
    def row_count(self):
        """
        If the task is finished, this value is the number of rows imported.
        If the task is not finished, this value is the number of rows parsed.
        """
        return self._row_count

    @property
    def state(self):
        return self.state_2_state.get(self._state, BulkInsertState.ImportUnknownState)

    @property
    def state_name(self) -> str:
        return self.state_2_name.get(self._state, "unknown state")

    @property
    def id_ranges(self):
        """
        auto generated id ranges if the primary key is auto generated

        the id list of response is id ranges
        for example, if the response return [1, 100, 200, 250]
        the full id list should be [1, 2, 3 ... , 99, 100, 200, 201, 202 ... , 249, 250]
        """
        return self._id_ranges

    @property
    def ids(self):
        """
        auto generated ids if the primary key is auto generated

        the id list of response is id ranges
        for example, if the response return [1, 100, 200, 250]
        the full id list should be [1, 2, 3 ... , 99, 100, 200, 201, 202 ... , 249, 250]
        """

        if len(self._id_ranges) % 2 != 0:
            raise AutoIDException(message=ExceptionsMessage.AutoIDIllegalRanges)

        ids = []
        for i in range(int(len(self._id_ranges) / 2)):
            begin = self._id_ranges[i * 2]
            end = self._id_ranges[i * 2 + 1] + 1
            for j in range(begin, end):
                ids.append(j)

        return ids

    @property
    def infos(self):
        """more informations about the task, progress percentage, file path, failed reason, etc."""
        return self._infos

    @property
    def failed_reason(self):
        """failed reason of the bulk insert task."""
        return self._infos.get(BulkInsertState.FAILED_REASON, "")

    @property
    def files(self):
        """data files of the bulk insert task."""
        return self._infos.get(BulkInsertState.IMPORT_FILES, "")

    @property
    def collection_name(self):
        """target collection's name of the bulk insert task."""
        return self._infos.get(BulkInsertState.IMPORT_COLLECTION, "")

    @property
    def partition_name(self):
        """target partition's name of the bulk insert task."""
        return self._infos.get(BulkInsertState.IMPORT_PARTITION, "")

    @property
    def create_timestamp(self):
        """the integer timestamp when this task is created."""
        return self._create_ts

    @property
    def create_time_str(self):
        """A readable string converted from the timestamp when this task is created."""
        ts = time.localtime(self._create_ts)
        return time.strftime("%Y-%m-%d %H:%M:%S", ts)

    @property
    def progress(self):
        """working progress percent value."""
        percent = self._infos.get(BulkInsertState.IMPORT_PROGRESS, "0")
        return int(percent)


class GrantItem:
    def __init__(self, entity):
        self._object = entity.object.name
        self._object_name = entity.object_name
        self._role_name = entity.role.name
        self._grantor_name = entity.grantor.user.name
        self._privilege = entity.grantor.privilege.name

    def __repr__(self) -> str:
        s = f"GrantItem: <object:{self.object}>, <object_name:{self.object_name}>, " \
            f"<role_name:{self.role_name}>, <grantor_name:{self.grantor_name}>, " \
            f"<privilege:{self.privilege}>"
        return s

    @property
    def object(self):
        return self._object

    @property
    def object_name(self):
        return self._object_name

    @property
    def role_name(self):
        return self._role_name

    @property
    def grantor_name(self):
        return self._grantor_name

    @property
    def privilege(self):
        return self._privilege


class GrantInfo:
    """
    GrantInfo groups:
    - GrantItem: <object:Collection>, <object_name:foo>, <role_name:x>, <grantor_name:root>, <privilege:Load>
    - GrantItem: <object:Global>, <object_name:*>, <role_name:x>, <grantor_name:root>, <privilege:CreateCollection>
    """

    def __init__(self, entities):
        groups = []
        for entity in entities:
            if isinstance(entity, milvus_types.GrantEntity):
                groups.append(GrantItem(entity))

        self._groups = groups

    def __repr__(self) -> str:
        s = "GrantInfo groups:"
        for g in self.groups:
            s += f"\n- {g}"
        return s

    @property
    def groups(self):
        return self._groups


class UserItem:
    def __init__(self, username, entities):
        self._username = username
        roles = []
        for entity in entities:
            if isinstance(entity, milvus_types.RoleEntity):
                roles.append(entity.name)
        self._roles = tuple(roles)

    def __repr__(self) -> str:
        s = f"UserItem: <username:{self.username}>, <roles:{self.roles}>"
        return s

    @property
    def username(self):
        return self._username

    @property
    def roles(self):
        return self._roles


class UserInfo:
    """
    UserInfo groups:
    - UserItem: <username:root>, <roles:('admin', 'public')>
    """

    def __init__(self, results):
        groups = []
        for result in results:
            if isinstance(result, milvus_types.UserResult):
                groups.append(UserItem(result.user.name, result.roles))

        self._groups = groups

    def __repr__(self) -> str:
        s = "UserInfo groups:"
        for g in self.groups:
            s += f"\n- {g}"
        return s

    @property
    def groups(self):
        return self._groups


class RoleItem:
    def __init__(self, role_name, entities):
        self._role_name = role_name
        users = []
        for entity in entities:
            if isinstance(entity, milvus_types.UserEntity):
                users.append(entity.name)
        self._users = tuple(users)

    def __repr__(self) -> str:
        s = f"RoleItem: <role_name:{self.role_name}>, <users:{self.users}>"
        return s

    @property
    def role_name(self):
        return self._role_name

    @property
    def users(self):
        return self._users


class RoleInfo:
    """
    RoleInfo groups:
    - UserItem: <role_name:admin>, <users:('root',)>
    """

    def __init__(self, results):
        groups = []
        for result in results:
            if isinstance(result, milvus_types.RoleResult):
                groups.append(RoleItem(result.role.name, result.users))

        self._groups = groups

    def __repr__(self) -> str:
        s = "RoleInfo groups:"
        for g in self.groups:
            s += f"\n- {g}"
        return s

    @property
    def groups(self):
        return self._groups


class ResourceGroupInfo:
    def __init__(self, resource_group):
        self._name = resource_group.name
        self._capacity = resource_group.capacity
        self._num_available_node = resource_group.num_available_node
        self._num_loaded_replica = resource_group.num_loaded_replica
        self._num_outgoing_node = resource_group.num_outgoing_node
        self._num_incoming_node = resource_group.num_incoming_node

    def __repr__(self) -> str:
        s = f"""ResourceGroupInfo:
<name:{self.name}>,
<capacity:{self.capacity}>,
<num_available_node:{self.num_available_node}>,
<num_loaded_replica:{self.num_loaded_replica}>,
<num_outgoing_node:{self.num_outgoing_node}>,
<num_incoming_node:{self.num_incoming_node}>"""
        return s


    @property
    def name(self):
        return self._name


    @property
    def capacity(self):
        return self._capacity


    @property
    def num_available_node(self):
        return self._num_available_node


    @property
    def num_loaded_replica(self):
        return self._num_loaded_replica


    @property
    def num_outgoing_node(self):
        return self._num_outgoing_node


    @property
    def num_incoming_node(self):
        return self._num_incoming_node



# Copyright (C) 2019-2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

import os
import copy
import re
import threading
from urllib import parse
from typing import Tuple

from ..client.check import is_legal_host, is_legal_port, is_legal_address
from ..client.grpc_handler import GrpcHandler

from .default_config import DefaultConfig, ENV_CONNECTION_CONF
from ..exceptions import ExceptionsMessage, ConnectionConfigException, ConnectionNotExistException


def synchronized(func):
    """
    Decorator in order to achieve thread-safe singleton class.
    """
    func.__lock__ = threading.Lock()

    def lock_func(*args, **kwargs):
        with func.__lock__:
            return func(*args, **kwargs)

    return lock_func


class SingleInstanceMetaClass(type):
    instance = None

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(cls, *args, **kwargs):
        if cls.instance:
            return cls.instance
        cls.instance = cls.__new__(cls)
        cls.instance.__init__(*args, **kwargs)
        return cls.instance

    @synchronized
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)


class Connections(metaclass=SingleInstanceMetaClass):
    """ Class for managing all connections of milvus.  Used as a singleton in this module.  """

    def __init__(self):
        """ Constructs a default milvus alias config

            default config will be read from env: MILVUS_DEFAULT_CONNECTION，
            or "localhost:19530"

        """
        self._alias = {}
        self._connected_alias = {}

        self.add_connection(default=self._read_default_config_from_os_env())

    def _read_default_config_from_os_env(self):
        """ Read default connection config from environment variable: MILVUS_DEFAULT_CONNECTION.
        Format is:
            [<user>@]host[:<port>]

            protocol is one of: http, https, tcp, or <empty>
        Examples::
            localhost
            localhost:19530
            test_user@localhost:19530
        """

        # no need to adjust http://xxx, https://xxx, tcp://xxxx,
        # because protocol is ignored
        # @see __generate_address

        conf = os.getenv(ENV_CONNECTION_CONF, "").strip()
        if not conf:
            conf = DefaultConfig.DEFAULT_HOST

        rex = re.compile(r"^(?:([^\s/\\:]+)@)?([^\s/\\:]+)(?::(\d{1,5}))?$")
        matched = rex.search(conf)

        if not matched:
            raise ConnectionConfigException(message=ExceptionsMessage.EnvConfigErr % (ENV_CONNECTION_CONF, conf))

        user, host, port = matched.groups()
        user = user or ""
        port = port or DefaultConfig.DEFAULT_PORT

        return {
            "user": user,
            "address": f"{host}:{port}"
        }

    def add_connection(self, **kwargs):
        """ Configures a milvus connection.

        Addresses priority in kwargs: address, uri, host and port

        :param kwargs:
            * *address* (``str``) -- Optional. The actual address of Milvus instance.
                Example address: "localhost:19530"

            * *uri* (``str``) -- Optional. The uri of Milvus instance.
                Example uri: "http://localhost:19530", "tcp:localhost:19530", "https://ok.s3.south.com:19530".

            * *host* (``str``) -- Optional. The host of Milvus instance.
                Default at "localhost", PyMilvus will fill in the default host if only port is provided.

            * *port* (``str/int``) -- Optional. The port of Milvus instance.
                Default at 19530, PyMilvus will fill in the default port if only host is provided.

        Example::

            connections.add_connection(
                default={"host": "localhost", "port": "19530"},
                dev1={"host": "localhost", "port": "19531"},
                dev2={"uri": "http://random.com/random"},
                dev3={"uri": "http://localhost:19530"},
                dev4={"uri": "tcp://localhost:19530"},
                dev5={"address": "localhost:19530"},
                prod={"uri": "http://random.random.random.com:19530"},
            )
        """
        for alias, config in kwargs.items():
            addr = self.__get_full_address(
                config.get("address", ""),
                config.get("uri", ""),
                config.get("host", ""),
                config.get("port", ""))

            if alias in self._connected_alias:
                if self._alias[alias].get("address") != addr:
                    raise ConnectionConfigException(message=ExceptionsMessage.ConnDiffConf % alias)

            alias_config = {
                "address": addr,
                "user": config.get("user", ""),
            }

            self._alias[alias] = alias_config

    def __get_full_address(self, address: str = "", uri: str = "", host: str = "", port: str = "") -> str:
        if address != "":
            if not is_legal_address(address):
                raise ConnectionConfigException(message=f"Illegal address: {address}, should be in form 'localhost:19530'")
        else:
            address = self.__generate_address(uri, host, port)

        return address

    def __generate_address(self, uri: str, host: str, port: str) -> str:
        illegal_uri_msg = "Illegal uri: [{}], should be in form 'http://example.com' or 'tcp://6.6.6.6:12345'"
        if uri != "":
            try:
                parsed_uri = parse.urlparse(uri)
            except (Exception) as e:
                raise ConnectionConfigException(message=f"{illegal_uri_msg.format(uri)}: <{type(e).__name__}, {e}>") from None

            if len(parsed_uri.netloc) == 0:
                raise ConnectionConfigException(message=illegal_uri_msg.format(uri))

            addr = parsed_uri.netloc if ":" in parsed_uri.netloc else f"{parsed_uri.netloc}:{DefaultConfig.DEFAULT_PORT}"
            if not is_legal_address(addr):
                raise ConnectionConfigException(message=illegal_uri_msg.format(uri))
            return addr

        host = host if host != "" else DefaultConfig.DEFAULT_HOST
        port = port if port != "" else DefaultConfig.DEFAULT_PORT

        if not is_legal_host(host):
            raise ConnectionConfigException(message=ExceptionsMessage.HostType)
        if not is_legal_port(port):
            raise ConnectionConfigException(message=ExceptionsMessage.PortType)
        if not 0 <= int(port) < 65535:
            raise ConnectionConfigException(message=f"port number {port} out of range, valid range [0, 65535)")

        return f"{host}:{port}"

    def disconnect(self, alias: str):
        """ Disconnects connection from the registry.

        :param alias: The name of milvus connection
        :type alias: str
        """
        if not isinstance(alias, str):
            raise ConnectionConfigException(message=ExceptionsMessage.AliasType % type(alias))

        if alias in self._connected_alias:
            self._connected_alias.pop(alias).close()

    def remove_connection(self, alias: str):
        """ Removes connection from the registry.

        :param alias: The name of milvus connection
        :type alias: str
        """
        if not isinstance(alias, str):
            raise ConnectionConfigException(message=ExceptionsMessage.AliasType % type(alias))

        self.disconnect(alias)
        self._alias.pop(alias, None)

    def connect(self, alias=DefaultConfig.DEFAULT_USING, user="", password="", **kwargs):
        """
        Constructs a milvus connection and register it under given alias.

        :param alias: The name of milvus connection
        :type  alias: str

        :param kwargs:
            * *address* (``str``) -- Optional. The actual address of Milvus instance.
                Example address: "localhost:19530"

            * *uri* (``str``) -- Optional. The uri of Milvus instance.
                Example uri: "http://localhost:19530", "tcp:localhost:19530", "https://ok.s3.south.com:19530".

            * *host* (``str``) -- Optional. The host of Milvus instance.
                Default at "localhost", PyMilvus will fill in the default host if only port is provided.

            * *port* (``str/int``) -- Optional. The port of Milvus instance.
                Default at 19530, PyMilvus will fill in the default port if only host is provided.

            * *user* (``str``) --
                Optional. Use which user to connect to Milvus instance. If user and password
                are provided, we will add related header in every RPC call.
            * *password* (``str``) --
                Optional and required when user is provided. The password corresponding to
                the user.
            * *secure* (``bool``) --
                Optional. Default is false. If set to true, tls will be enabled.
            * *client_key_path* (``str``) --
                Optional. If use tls two-way authentication, need to write the client.key path.
            * *client_pem_path* (``str``) --
                Optional. If use tls two-way authentication, need to write the client.pem path.
            * *ca_pem_path* (``str``) --
                Optional. If use tls two-way authentication, need to write the ca.pem path.
            * *server_pem_path* (``str``) --
                Optional. If use tls one-way authentication, need to write the server.pem path.
            * *server_name* (``str``) --
                Optional. If use tls, need to write the common name.

        :raises NotImplementedError: If handler in connection parameters is not GRPC.
        :raises ParamError: If pool in connection parameters is not supported.
        :raises Exception: If server specified in parameters is not ready, we cannot connect to
                           server.

        :example:
            >>> from pymilvus import connections
            >>> connections.connect("test", host="localhost", port="19530")
        """
        if not isinstance(alias, str):
            raise ConnectionConfigException(message=ExceptionsMessage.AliasType % type(alias))

        def connect_milvus(**kwargs):
            gh = GrpcHandler(**kwargs)

            t = kwargs.get("timeout")
            timeout = t if isinstance(t, (int, float)) else DefaultConfig.DEFAULT_CONNECT_TIMEOUT

            gh._wait_for_channel_ready(timeout=timeout)
            kwargs.pop('password')
            kwargs.pop('secure', None)

            self._connected_alias[alias] = gh
            self._alias[alias] = copy.deepcopy(kwargs)

        def with_config(config: Tuple) -> bool:
            for c in config:
                if c != "":
                    return True

            return False

        config = (
            kwargs.pop("address", ""),
            kwargs.pop("uri", ""),
            kwargs.pop("host", ""),
            kwargs.pop("port", "")
        )

        if with_config(config):
            in_addr = self.__get_full_address(*config)
            kwargs["address"] = in_addr

            if self.has_connection(alias):
                if self._alias[alias].get("address") != in_addr:
                    raise ConnectionConfigException(message=ExceptionsMessage.ConnDiffConf % alias)

            connect_milvus(**kwargs, user=user, password=password)

        else:
            if alias not in self._alias:
                raise ConnectionConfigException(message=ExceptionsMessage.ConnLackConf % alias)

            connect_alias = dict(self._alias[alias].items())
            connect_alias["user"] = user
            connect_milvus(**connect_alias, password=password, **kwargs)

    def list_connections(self) -> list:
        """ List names of all connections.

        :return list:
            Names of all connections.

        :example:
            >>> from pymilvus import connections
            >>> connections.connect("test", host="localhost", port="19530")
            >>> connections.list_connections()
            // TODO [('default', None), ('test', <pymilvus.client.grpc_handler.GrpcHandler object at 0x7f05003f3e80>)]
        """
        return [(k, self._connected_alias.get(k, None)) for k in self._alias]

    def get_connection_addr(self, alias: str):
        """
        Retrieves connection configure by alias.

        :param alias: The name of milvus connection
        :type  alias: str

        :return dict:
            The connection configure which of the name is alias.
            If alias does not exist, return empty dict.

        :example:
            >>> from pymilvus import connections
            >>> connections.connect("test", host="localhost", port="19530")
            >>> connections.list_connections()
            [('default', None), ('test', <pymilvus.client.grpc_handler.GrpcHandler object at 0x7f4045335f10>)]
            >>> connections.get_connection_addr('test')
            {'host': 'localhost', 'port': '19530'}
        """
        if not isinstance(alias, str):
            raise ConnectionConfigException(message=ExceptionsMessage.AliasType % type(alias))

        return self._alias.get(alias, {})

    def has_connection(self, alias: str) -> bool:
        """ Check if connection named alias exists.

        :param alias: The name of milvus connection
        :type  alias: str

        :return bool:
            if the connection of name alias exists.

        :example:
            >>> from pymilvus import connections
            >>> connections.connect("test", host="localhost", port="19530")
            >>> connections.list_connections()
            [('default', None), ('test', <pymilvus.client.grpc_handler.GrpcHandler object at 0x7f4045335f10>)]
            >>> connections.get_connection_addr('test')
            {'host': 'localhost', 'port': '19530'}
        """
        if not isinstance(alias, str):
            raise ConnectionConfigException(message=ExceptionsMessage.AliasType % type(alias))
        return alias in self._connected_alias

    def _fetch_handler(self, alias=DefaultConfig.DEFAULT_USING) -> GrpcHandler:
        """ Retrieves a GrpcHandler by alias. """
        if not isinstance(alias, str):
            raise ConnectionConfigException(message=ExceptionsMessage.AliasType % type(alias))

        conn = self._connected_alias.get(alias, None)
        if conn is None:
            raise ConnectionNotExistException(message=ExceptionsMessage.ConnectFirst)

        return conn


# Singleton Mode in Python
connections = Connections()



import grpc
import grpc_testing
import pytest
import random

from pymilvus.grpc_gen import milvus_pb2, schema_pb2, common_pb2
from pymilvus import Milvus, DataType


class Fields:
    class NormalizedField:
        def __init__(self, **kwargs):
            self.name = kwargs.get("name", None)
            self.is_primary_key = kwargs.get("is_primary_key", False)
            self.data_type = kwargs.get("data_type", None)
            self.type_params = kwargs.get("type_params", dict())
            self.autoID = kwargs.get("autoID", False)

        def __eq__(self, other):
            if isinstance(other, Fields.NormalizedField):
                return self.name == other.name and \
                       self.is_primary_key == other.is_primary_key and \
                       self.data_type == other.data_type and \
                       self.type_params == other.type_params and \
                       self.autoID == other.autoID
            return False

        def __repr__(self):
            dump = f"(name: {self.name}"
            dump += f", id_primary_key:{self.is_primary_key}"
            dump += f", data_type:{self.data_type}"
            dump += f", type_params:{self.type_params}"
            dump += f", autoID:{self.autoID})"
            return dump

    @classmethod
    def equal(cls, grpc_fields, dict_fields):
        n_grpc_fields = {
            field.name: Fields.NormalizedField(name=field.name,
                                               is_primary_key=field.is_primary_key,
                                               data_type=field.data_type,
                                               type_params={pair.key: pair.value for pair in field.type_params},
                                               autoID=field.autoID
                                               )
            for field in grpc_fields}
        n_dict_fields = {
            field["name"]: Fields.NormalizedField(name=field["name"],
                                                  is_primary_key=field.get("is_primary", False),
                                                  data_type=field["type"],
                                                  type_params=field.get("params", dict()),
                                                  autoID=field.get("auto_id", False)
                                                  )
            for field in dict_fields
        }
        return n_grpc_fields == n_dict_fields


class TestCreateCollection:
    @pytest.fixture(scope="function")
    def collection_name(self):
        return f"test_collection_{random.randint(100000, 999999)}"

    def setup(self) -> None:
        self._real_time = grpc_testing.strict_real_time()
        self._real_time_channel = grpc_testing.channel(
            milvus_pb2.DESCRIPTOR.services_by_name.values(), self._real_time)
        self._servicer = milvus_pb2.DESCRIPTOR.services_by_name['MilvusService']
        self._milvus = Milvus(channel=self._real_time_channel)

    def teardown(self) -> None:
        pass

    def test_create_collection(self, collection_name):
        id_field = {
            "name": "my_id",
            "type": DataType.INT64,
            "auto_id": True,
            "is_primary": True,
        }
        vector_field = {
            "name": "embedding",
            "type": DataType.FLOAT_VECTOR,
            "metric_type": "L2",
            "params": {"dim": "4"},
        }
        fields = {"fields": [id_field, vector_field]}
        future = self._milvus.create_collection(collection_name=collection_name, fields=fields, _async=True)

        invocation_metadata, request, rpc = self._real_time_channel.take_unary_unary(
            self._servicer.methods_by_name['CreateCollection']
        )
        rpc.send_initial_metadata(())
        rpc.terminate(common_pb2.Status(error_code=common_pb2.Success, reason="success"), (), grpc.StatusCode.OK, '')

        request_schema = schema_pb2.CollectionSchema()
        request_schema.ParseFromString(request.schema)

        assert request.collection_name == collection_name
        assert Fields.equal(request_schema.fields, fields["fields"])

        return_value = future.result()
        assert return_value.error_code == common_pb2.Success
        assert return_value.reason == "success"



# Copyright (C) 2019-2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.
import copy

import numpy
import pandas

from ..exceptions import DataNotMatchException, DataTypeNotSupportException, ExceptionsMessage


class Prepare:
    @classmethod
    def prepare_insert_data(cls, data, schema):
        if not isinstance(data, (list, tuple, pandas.DataFrame)):
            raise DataTypeNotSupportException(message=ExceptionsMessage.DataTypeNotSupport)

        fields = schema.fields
        entities = []  # Entities
        raw_lengths = []  # Check if each row has the same numbers.

        if isinstance(data, pandas.DataFrame):
            if schema.auto_id:
                if schema.primary_field.name in data:
                    if len(fields) != len(data.columns):
                        raise DataNotMatchException(message=ExceptionsMessage.FieldsNumInconsistent)
                    if not data[schema.primary_field.name].isnull().all():
                        raise DataNotMatchException(message=ExceptionsMessage.AutoIDWithData)
                else:
                    if len(fields) != len(data.columns) + 1:
                        raise DataNotMatchException(message=ExceptionsMessage.FieldsNumInconsistent)
            else:
                if len(fields) != len(data.columns):
                    raise DataNotMatchException(message=ExceptionsMessage.FieldsNumInconsistent)
            for i, field in enumerate(fields):
                if field.is_primary and field.auto_id:
                    continue
                entities.append({"name": field.name,
                                 "type": field.dtype,
                                 "values": list(data[field.name])})
                raw_lengths.append(len(data[field.name]))
        else:
            if schema.auto_id and len(data) != len(fields) - 1:
                raise DataNotMatchException(message=ExceptionsMessage.FieldsNumInconsistent)

            tmp_fields = copy.deepcopy(fields)
            for i, field in enumerate(tmp_fields):
                #  TODO Goose: Checking auto_id and is_primary only, maybe different than
                #  schema.is_primary, schema.auto_id, need to check why and how schema is built.
                if field.is_primary and field.auto_id:
                    tmp_fields.pop(i)

            for i, field in enumerate(tmp_fields):
                # TODO: check string.
                if isinstance(data[i], numpy.ndarray):
                    data[i] = data[i].tolist()

                entities.append({
                    "name": field.name,
                    "type": field.dtype,
                    "values": data[i]})
                raw_lengths.append(len(data[i]))

        # TODO Goose: check correctness AFTER copy is too expensive.
        lengths = list(set(raw_lengths))
        if len(lengths) > 1:
            raise DataNotMatchException(message=ExceptionsMessage.DataLengthsInconsistent)

        return entities
